import datetime
from string import ascii_lowercase, ascii_uppercase
#from types import NoneType

import dateutil
import numpy as np
from ase import Atoms

from colabfit import ATOMS_LABELS_FIELD, ATOMS_NAME_FIELD
from colabfit.tools.schema import config_schema
from colabfit.tools.utilities import (
    _empty_dict_from_schema,
    _hash,
    _parse_unstructured_metadata,
)


class AtomicConfiguration(Atoms):
    """
    An AtomicConfiguration is an extension of a :class:`BaseConfiguration` and an
    :class:`ase.Atoms` object that is guaranteed to have the following fields in
    its :attr:`info` dictionary:

    - :attr:`~colabfit.ATOMS_NAME_FIELD` = :code:"_name"
    - :attr:`~colabfit.ATOMS_LABELS_FIELD` = :code:"_labels"
    """

    def __init__(
        self,
        co_md_map=None,
        info=None,
        **kwargs,
    ):
        """
        Constructs an AtomicConfiguration. Calls :meth:`BaseConfiguration.__init__()`
        and :meth:`ase.Atoms.__init__()`

        Args:
            co_md_map (dict) optional:
                Property map of metadata to be used to set configuration metadata for a
                configuration. This should be called at creation of AtomicConfiguration
                in order to include metadata in the hash.
            **kwargs:
                Other keyword arguments that can be passed to
                :meth:`ase.Atoms.__init__()`
        """
        if ATOMS_NAME_FIELD in info:
            names = info[ATOMS_NAME_FIELD]
        else:
            names = None
        if "atomic_numbers" in list(kwargs.keys()):
            kwargs["numbers"] = kwargs.pop("atomic_numbers")

        Atoms.__init__(self, **kwargs)
        self._array_order = np.lexsort(
            (
                self.arrays["positions"][:, 2],
                self.arrays["positions"][:, 1],
                self.arrays["positions"][:, 0],
            )
        )
        self.unique_identifier_kw = [
            "atomic_numbers",
            "positions",
            "cell",
            "pbc",
            #"metadata_id",
        ]
        #self.unique_identifier_kw.extend([f"positions_{i:02d}" for i in range(1, 20)])
        self.info = info
        self.metadata = self.set_metadata(co_md_map)
        if isinstance(names, str):
            self.names = [names]
        else:
            self.names = names

        labels = self.info.pop(ATOMS_LABELS_FIELD, None)
        if isinstance(labels, str):
            self.labels = [labels]
        #elif not isinstance(labels, list):
        #    raise TypeError("Labels must be a string or a list of strings or None")
        else:
            self.labels = labels
        self.spark_row = self.to_spark_row()
        self._hash = hash(self)
        self.id = f"CO_{self._hash}"

        self.spark_row["id"] = self.id
        self.spark_row["hash"] = str(self._hash)
        # self.spark_row["dataset_ids"] = [self.dataset_id]
        self.spark_row = self.spark_row
        # Check for name conflicts in info/arrays; would cause bug in parsing
        if set(self.info.keys()).intersection(set(self.arrays.keys())):
            raise RuntimeError(
                "The same key should not be used in both Configuration.info and "
                "Configuration.arrays"
            )

    def set_metadata(self, co_md_map):
        """
        Returns metadata for a configuration from a property map.
        This should be called at creation of AtomicConfiguration in order
        to include metadata in the hash.

        Args:
            co_md_map (dict): Property map of metadata to be used to set configuration
            metadata for a configuration.

        """
        if co_md_map is None:
            co_md_map = {}
        gathered_fields = {}
        for md_field in co_md_map.keys():
            if "value" in co_md_map[md_field]:
                v = co_md_map[md_field]["value"]
            elif "field" in co_md_map[md_field]:
                field_key = co_md_map[md_field]["field"]

                if field_key in self.info:
                    v = self.info[field_key]
                elif field_key in self.arrays:
                    v = self.arrays[field_key]
                else:
                    # No keys are required; ignored if missing
                    continue
            else:
                # No keys are required; ignored if missing
                continue

            if "units" in co_md_map[md_field]:
                gathered_fields[md_field] = {
                    f"{md_field}": v,
                    f"{md_field}_unit": co_md_map[md_field]["units"],
                }
            else:
                gathered_fields[md_field] = v

        return _parse_unstructured_metadata(gathered_fields)

    def configuration_summary(self):
        """Extracts useful metadata from a Configuration

        Gathers the following information from a Configuration:

        * :code:`nsites`: the total number of atoms
        * :code:`nelements`: the total number of unique element types
        * :code:`elements`: the element types
        * :code:`elements_ratios`: elemental ratio of the species
        * :code:`chemical_formula_reduced`: the reduced chemical formula
        * :code:`chemical_formula_anonymous`: the chemical formula
        * :code:`chemical_formula_hill`: the hill chemical formulae
        * :code:`nperiodic_dimensions`: the numbers of periodic dimensions
        * :code:`dimension_types`: the periodic boundary condition

        Returns:
            dict: Keys and their associated values that will be included under a
            Configuration's entry in the Database
        """

        atomic_species = self.get_chemical_symbols()

        natoms = len(atomic_species)
        elements = sorted(list(set(atomic_species)))
        nelements = len(elements)
        elements_ratios = [atomic_species.count(el) / natoms for el in elements]

        species_counts = [atomic_species.count(sp) for sp in elements]

        # Build per-element proportions
        from functools import reduce
        from math import gcd

        def find_gcd(_list):
            x = reduce(gcd, _list)
            return x

        count_gcd = find_gcd(species_counts)
        species_proportions = [sc // count_gcd for sc in species_counts]

        chemical_formula_reduced = ""
        for elem, elem_prop in zip(elements, species_proportions):
            chemical_formula_reduced += "{}{}".format(
                elem, "" if elem_prop == 1 else str(elem_prop)
            )

        # Replace elements with A, B, C, ...
        species_proportions = sorted(species_proportions, reverse=True)

        chemical_formula_anonymous = ""
        for spec_idx, spec_count in enumerate(species_proportions):
            # OPTIMADE uses A...Z, then Aa..Za, ..., up to Az...Zz

            count1 = spec_idx // 26
            count2 = spec_idx % 26

            if count1 == 0:
                anon_spec = ascii_uppercase[count2]
            else:
                anon_spec = ascii_uppercase[count1] + ascii_lowercase[count2]

            chemical_formula_anonymous += anon_spec
            if spec_count > 1:
                chemical_formula_anonymous += str(spec_count)

        species = []
        for el in elements:
            # https://github.com/Materials-Consortia/OPTIMADE/blob/develop/optimade.rst#7214species
            species.append(
                {
                    "name": el,
                    "chemical_symbols": [el],
                    "concentration": [1.0],
                }
            )

        return {
            "nsites": natoms,
            "elements": elements,
            "nelements": nelements,
            "elements_ratios": elements_ratios,
            "chemical_formula_anonymous": chemical_formula_anonymous,
            "chemical_formula_reduced": chemical_formula_reduced,
            "chemical_formula_hill": self.get_chemical_formula(),
            "dimension_types": self.get_pbc().astype(int).tolist(),
            "nperiodic_dimensions": int(sum(self.get_pbc())),
            # 'species': species,  # Is this ever used?
        }

    def set_dataset_id(self, dataset_id):
        self.dataset_id = dataset_id
        self.spark_row["dataset_ids"] = [dataset_id]

    def to_spark_row(self):
        co_dict = _empty_dict_from_schema(config_schema)
        co_dict["cell"] = self.cell.array.astype(float).tolist()
        co_dict["positions"] = self.positions.astype(float).tolist()
        co_dict["names"] = self.names
        co_dict["labels"] = self.labels
        co_dict["pbc"] = self.pbc.astype(bool).tolist()
        co_dict["last_modified"] = dateutil.parser.parse(
            datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        )
        co_dict["atomic_numbers"] = self.numbers.astype(int).tolist()
        #if self.metadata is not None:
        #    co_dict.update(self.metadata)
        co_dict.update(self.configuration_summary())
        return co_dict

    @classmethod
    def from_ase(self, atoms, co_md_map=None):
        """
        Generates an :class:`AtomicConfiguration` from an :code:`ase.Atoms` object.
        """
        # Workaround for bug in todict() fromdict() with constraints.
        # Merge request: https://gitlab.com/ase/ase/-/merge_requests/2574
        # This means kwargs need to be same as those in ASE
        dct = atoms.todict()
        kw = {name: dct.pop(name) for name in ["numbers", "positions", "cell", "pbc"]}
        #  Round positions (and maybe cell) to some decimal place so that configurations hash the same
        kw['positions'] = np.round(kw['positions'],8)
        kw['cell'] = np.round(kw['cell'],8)
        constraints = dct.pop("constraints", None)
        if constraints:
            constraints = [c.todict() for c in atoms.constraints]
            from ase.constraints import dict2constraint

            constraints = [dict2constraint(d) for d in constraints]
        info = dct.pop("info", None)
        for k, v in info.items():
            if k in [ATOMS_NAME_FIELD, ATOMS_LABELS_FIELD]:
                if not isinstance(v, set):
                    if not isinstance(v, list):
                        v = [v]
                    info[k] = list(set(v))
                else:
                    info[k] = list(v)
            else:
                info[k] = v
        config = self(
            constraint=constraints,
            celldisp=dct.pop("celldisp", None),
            info=info,
            co_md_map=co_md_map,
            **kw,
        )
        natoms = len(atoms)
        for name, arr in dct.items():
            assert len(arr) == natoms, name
            assert isinstance(arr, np.ndarray)
            config.arrays[name] = arr
        # for k, v in atoms.arrays.items():
        #     conf.arrays[k] = v

        return config

    @staticmethod
    def aggregate_configuration_summaries(db, co_hashes, verbose=False):
        """
        Gathers the following information from a collection of Configurations:

        * :code:`nconfigurations`: the total number of configurations
        * :code:`nsites`: the total number of atoms
        * :code:`nelements`: the total number of unique element types
        * :code:`elements`: the element types
        # * :code:`individual_elements_ratios`: a set of elements ratios generated
        #   by looping over each configuration, extracting its concentration of
        #   each element, and adding the tuple of concentrations to the set
        * :code:`total_elements_ratios`: the ratio of the total count of atoms
        of each element type over :code:`nsites`
        * :code:`chemical_formula_reduced`: the set of all reduced chemical
        formulae
        * :code:`chemical_formula_anonymous`: the set of all anonymous chemical
        formulae
        * :code:`chemical_formula_hill`: the set of all hill chemical formulae
        * :code:`nperiodic_dimensions`: the set of all numbers of periodic
        dimensions
        * :code:`dimension_types`: the set of all periodic boundary choices

        Args:
            db (:code:`MongoDatabase` object):
                Database client in which to search for dataset-ID
            co_hashes (list of str):
                List of hashes of configurations to aggregate. /
            verbose (bool, default=False):
                If True, prints a progress bar

        Returns:
            dict: Aggregated Configuration information
        """
        return NotImplementedError

    def __str__(self):
        ase_str = super().__str__()
        return "AtomicConfiguration(name={}, {})".format(
            self.info[ATOMS_NAME_FIELD], ase_str[20:-1]
        )

    def __hash__(self):
        return _hash(self.spark_row, sorted(self.unique_identifier_kw), False)



from html.parser import HTMLParser


class InvalidHeaderError(Exception):
    pass


class MarkdownFormatError(Exception):
    pass


class BadTableFormatting(Exception):
    pass


class DatasetParser(HTMLParser):

    KNOWN_HEADERS = [
        'Name',
        'Authors',
        'Links',
        'Description',
        'Storage format',
        'Summary',
        'Properties',
        'Property settings',
        'Configuration sets',
        'Configuration labels',
    ]

    def __init__(self):
        super().__init__()

        self.data           = {}
        self._t             = None
        self._table         = None
        self._loading_table = False
        self._loading_row   = False
        self._header        = None
        self._href          = None

    def get_data(self, k):
        try:
            return self.data[k]
        except:
            raise MarkdownFormatError(
                f"An error occurred trying to access DatasetParser.data['{k}']"
            )


    def handle_starttag(self, tag, attrs):
        self._t = tag

        if tag == 'table':
            self._table = []
            self._loading_table = True
        elif (tag == 'thead') or ('tag' == 'tbody'):
            pass
        elif tag == 'tr':
            self._loading_row = True
            self._table.append([])

        for att in attrs:
            if att[0] == 'href':
                self._href = att[1]


    def handle_endtag(self, tag):
        if tag == 'table':
            self.data[self._header] = self._table
            self._table = None
            self._loading_table = False
        elif tag == 'tr':
            self._loading_row = False

    def handle_data(self, data):
        data = data.rstrip('\n')

        if data:
            if self._t == 'h1':
                # Begin reading new block
                if data not in self.KNOWN_HEADERS:
                    # Unknown sections are ignored
                    return
                self._header = data
                self.data[self._header] = []

            else:
                # Add data to current block
                if not self._loading_table:
                    # Adding basic text
                    self.data[self._header] += [_.strip() for _ in data.split('\n')]
                else:
                    # Specifically adding to a table
                    if self._loading_row:
                        if self._href is not None:
                            # Hyperlinks in a table (often for files) are
                            # grouped by column to avoid indexing errors

                            if (len(self._table[-1]) == 0) or not isinstance(
                                self._table[-1][-1], list):
                                self._table[-1].append([(data, self._href)])
                            else:
                                self._table[-1][-1].append((data, self._href))

                            self._href = None
                        else:
                            if data != ', ':
                                # ', ' can occur in lists of files
                                self._table[-1].append(data)



import datetime
import itertools
import json
import os
import tempfile
import warnings
from collections import namedtuple
from copy import deepcopy

import dateutil.parser
import kim_edn
import numpy as np
from ase.units import create_units
from kim_property import (
    check_instance_optional_key_marked_required_are_present,
    kim_property_create,
)
from kim_property.definition import PROPERTY_ID as VALID_KIM_ID

from colabfit.tools.configuration import AtomicConfiguration
from colabfit.tools.schema import property_object_schema
from colabfit.tools.utilities import (
    _empty_dict_from_schema,
    _hash,
    _parse_unstructured_metadata,
)

from kim_property.create import KIM_PROPERTIES

EDN_KEY_MAP = {
    "energy": "unrelaxed-potential-energy",
    "forces": "unrelaxed-potential-forces",
    "stress": "unrelaxed-cauchy-stress",
    "virial": "unrelaxed-cauchy-stress",
}

UNITS = create_units("2014")

UNITS["angstrom"] = UNITS["Ang"]
UNITS["bohr"] = UNITS["Bohr"]
UNITS["hartree"] = UNITS["Hartree"]
UNITS["rydberg"] = UNITS["Rydberg"]
UNITS["debye"] = UNITS["Debye"]
UNITS["kbar"] = UNITS["bar"] * 1000

prop_info = namedtuple("prop_info", ["key", "unit", "dtype"])
energy_info = prop_info("energy", "eV", float)
force_info = prop_info("forces", "eV/angstrom", list)
stress_info = prop_info("stress", "eV/angstrom^3", list)
MAIN_KEY_MAP = {
    "energy": energy_info,
    "atomic-forces": force_info,
    "cauchy-stress": stress_info,
    "atomization-energy": energy_info,
    "formation-energy": energy_info,
    "band-gap": energy_info,
}

# These are fields that are related to the geometry of the atomic structure
# or the OpenKIM Property Definition and shouldn't be used for equality checks
_ignored_fields = [
    "property-title",
    "property-description",
    "instance-id",
    "species",
    "unrelaxed-configuration-positions",
    "unrelaxed-periodic-cell-vector-1",
    "unrelaxed-periodic-cell-vector-2",
    "unrelaxed-periodic-cell-vector-3",
]
_hash_ignored_fields = [
    "id",
    "hash",
    "last_modified",
    "multiplicity",
    "metadata_path",
    "metadata_size",
]


def energy_to_schema(prop_name, en_prop: dict):
    if en_prop.get("energy") is None:
        return {}
    new_name = prop_name.replace("-", "_")
    en_dict = {
        f"{new_name}": en_prop["energy"]["source-value"],
        f"{new_name}_unit": en_prop["energy"]["source-unit"],
        f"{new_name}_per_atom": en_prop["per-atom"]["source-value"],
    }
    ref_en_dict = en_prop.get("reference-energy")
    if ref_en_dict is not None:
        ref_en = ref_en_dict["source-value"]
        ref_unit = ref_en_dict["source-unit"]
        en_dict[f"{new_name}_reference"] = ref_en
        en_dict[f"{new_name}_reference_unit"] = ref_unit

    return en_dict


def atomic_forces_to_schema(af_prop: dict):
    if af_prop.get("forces") is None:
        return {}
    af_dict = {
        "atomic_forces_00": af_prop["forces"]["source-value"],
        "atomic_forces_unit": af_prop["forces"]["source-unit"],
    }
    return af_dict


def cauchy_stress_to_schema(cs_prop: dict):
    if cs_prop.get("stress") is None:
        return {}
    cs_dict = {
        "cauchy_stress": cs_prop["stress"]["source-value"],
        "cauchy_stress_unit": cs_prop["stress"]["source-unit"],
        "cauchy_stress_volume_normalized": cs_prop["volume-normalized"]["source-value"],
    }
    return cs_dict


def band_gap_to_schema(bg_prop: dict):
    if bg_prop.get("energy") is None:
        return {}
    bg_dict = {
        "electronic_band_gap": bg_prop["energy"]["source-value"],
        "electronic_band_gap_unit": bg_prop["energy"]["source-unit"],
        "electronic_band_gap_type": bg_prop.get("type", {"source-value": "direct"})[
            "source-value"
        ],
    }
    return bg_dict


prop_to_row_mapper = {
    "energy": energy_to_schema,
    "atomic-forces": atomic_forces_to_schema,
    "cauchy-stress": cauchy_stress_to_schema,
    "band-gap": band_gap_to_schema,
}


def md_from_map(pmap_md, config: AtomicConfiguration) -> tuple:
    """
    Extract metadata from a property map.
    Returns metadata dict as a JSON string..
    """
    if isinstance(pmap_md, list):
        pmap_md = pmap_md[0]
    gathered_fields = {}
    for md_field in pmap_md.keys():
        if "value" in pmap_md[md_field]:
            v = pmap_md[md_field]["value"]
        elif "field" in pmap_md[md_field]:
            field_key = pmap_md[md_field]["field"]
            if md_field == 'metadata' or '_metadata':
                if field_key in config.info:
                    assert isinstance(config.info[field_key], dict)
                    for k2,v2 in config.info[field_key].items():
                       gathered_fields[k2] = v2
                continue
            if field_key in config.info:
                v = config.info[field_key]
            elif field_key in config.arrays:
                v = config.arrays[field_key]
            else:
                continue  # No keys are required; ignored if missing
        else:
            continue  # No keys are required; ignored if missing

        if "units" in pmap_md[md_field]:
            gathered_fields[md_field] = {
                "source-value": v,
                "source-unit": pmap_md[md_field]["units"],
            }
        else:
            gathered_fields[md_field] =  v
    return json.dumps(gathered_fields)


class PropertyParsingError(Exception):
    pass


class InvalidPropertyDefinition(Exception):
    pass


class Property(dict):
    """
    A Property is used to store the results of some kind of calculation or
    experiment, and should be mapped to an `OpenKIM Property Definition
    <https://openkim.org/doc/schema/properties-framework/>`_. Best
    practice is for the Property to also point to one or more
    PropertySettings objects that fully define the conditions under which the
    Property was obtained.

    Attributes:

        definition (dict):
            A KIM Property Definition

        instance (dict):
            A dictionary defining an OpenKIM Property Instance.
            For more details, see the `OpenKIM Property Framework
            <https://openkim.org/doc/schema/properties-framework/>`_
            documentation. In most cases, this dictionary should not be manually
            constructed by the user. Instead, a Property Definition and a
            Configuration should be passed to the :meth:`from_definition`
            function.

        property_map (dict):
            key = a string that can be used as a key like :code:`self.instance[key]`

            value = A sub-dictionary with the following keys:

            * :attr:`field`:
                A field name used to access :attr:`Configuration.info` or
                :attr:`Configuration.arrays`
            * :attr:`units`:
                A string matching one of the units names in
                `ase.units <https://wiki.fysik.dtu.dk/ase/ase/units.html>`_.
                These units will be used to convert the given units to eV, Angstrom,
                a.m.u., Kelvin, ... For compound units (e.g. "eV/Ang"), the string will
                be split on '*' and '/'. The raw data will be multiplied by the first
                unit and anything preceded by a '*'. It will be divided by anything
                preceded by a '/'.
    """

    _observers = []

    def __init__(
        self,
        definitions,
        instance,
        property_map=None,
        metadata=None,
        standardize_energy=True,
        nsites=None,
        dataset_id=None,
    ):
        """
        Args:
            definitions (list(dict)):
                KIM Property Definitions
            instance (dict):
                A dictionary defining an OpenKIM Property Instance
            property_map (dict):
                A property map as described in the Property attributes section.
            standardize_energy (bool):
                If True, converts units to those expected by ColabFit. Default
                is True
        """
        # self.unique_identifier_kw = [
        #     k
        #     for k in property_object_schema.fieldNames()
        #     if k not in _hash_ignored_fields
        # ]
        # TODO: dynamically set this
        #self.unique_identifier_kw = [
            #"adsorption_energy",
            #"atomic_forces_00",
            #"atomization_energy",
            #"cauchy_stress",
            #"cauchy_stress_volume_normalized",
            #"chemical_formula_hill",
            #"configuration_id",
            #"dataset_id",
            #"electronic_band_gap",
            #"electronic_band_gap_type",
            #"energy",
            #"formation_energy",
            #"metadata_id",
            #"method",
            #"software",
        #]
        #self.unique_identifier_kw.extend(
        #    [f"atomic_forces_{i:02d}" for i in range(1, 20)]
        #)
        self.instance = instance
        self.definitions = definitions
        self.nsites = nsites
        if property_map is not None:
            self.property_map = dict(property_map)
        else:
            self.property_map = {}
        self.metadata = metadata
        self.chemical_formula_hill = instance.pop("chemical_formula_hill")
        if standardize_energy:
            self.standardize_energy()
        if dataset_id is not None:
            self.dataset_id = dataset_id
        self.spark_row = self.to_spark_row()
        # TODO: Dynamically get unique_identifiers
        self.unique_identifier_kw = []
        for k,v in self.spark_row.items():
            if k not in ['last_modified']:
                self.unique_identifier_kw.append(k)
        self._hash = _hash(self.spark_row, self.unique_identifier_kw, False)
        self.spark_row["hash"] = str(self._hash)
        self._id = f"PO_{self._hash}"
        if len(self._id) > 28:
            self._id = self._id[:28]
        self.spark_row["id"] = self._id

    @property
    def instance(self):
        return self._instance

    @instance.setter
    def instance(self, edn):
        self._instance = edn
        fields = []
        for key in self._instance:
            if key in _ignored_fields:
                continue
            fields.append(key)
        self._property_fields = fields

    @property
    def property_fields(self):
        return self._property_fields

    @classmethod
    def get_kim_instance(
        cls,
        definition,
    ):
        """
        A function for constructing a Property given a property setting hash, a property
        definition, and a property map.

        Args:

            definition (dict):
                A valid KIM Property Definition

            configuration (AtomicConfiguration):
                An AtomicConfiguration object from which to extract the property data

            property_map (dict):
                A property map as described in the Property attributes section.

        """
        # global KIM_PROPERTIES
        load_from_existing = False
        if isinstance(definition, dict):
            # TODO: this is probably slowing things down a bit
            dummy_dict = deepcopy(definition)
            # Spoof if necessary
            if VALID_KIM_ID.match(dummy_dict["property-id"]) is None:
                # Invalid ID. Try spoofing it
                dummy_dict["property-id"] = "tag:@,0000-00-00:property/"
                dummy_dict["property-id"] += definition["property-id"]

            load_from_existing = dummy_dict["property-id"] in KIM_PROPERTIES
            definition_name = dummy_dict["property-id"]

        elif isinstance(definition, str):
            if os.path.isfile(definition):
                dummy_dict = kim_edn.load(definition)
                # Check if you need to spoof the property ID to trick OpenKIM
                if VALID_KIM_ID.match(dummy_dict["property-id"]) is None:
                    # Invalid ID. Try spoofing it
                    dummy_dict["property-id"] = (
                        "tag:@,0000-00-00:property/" + dummy_dict["property-id"]
                    )
                    warnings.warn(
                        "Invalid KIM property-id; Temporarily "
                        f"renaming to {dummy_dict['property-id']}"
                    )
                load_from_existing = dummy_dict["property-id"] in KIM_PROPERTIES
                definition_name = dummy_dict["property-id"]
            else:
                # Then this has to be an existing (or added) KIM Property Definition
                load_from_existing = True
                # It may have been added previously, but spoofed
                if VALID_KIM_ID.match(definition) is None:
                    # Invalid ID. Try spoofing it
                    definition_name = "tag:@,0000-00-00:property/" + definition
        else:
            raise InvalidPropertyDefinition(
                "Property definition must either be a dictionary "
                "or a path to an EDN file"
            )
        if load_from_existing:
            instance = kim_edn.loads(
                kim_property_create(
                    instance_id=1,
                    property_name=definition_name,
                )
            )[0]
        else:
            with tempfile.NamedTemporaryFile("w") as tmp:
                # Hack to avoid the fact that "property-name" has to be a dictionary
                # in order for OpenKIM's check_property_definition to work

                if "property-name" in dummy_dict:
                    del dummy_dict["property-name"]
                tmp.write(json.dumps(dummy_dict))
                tmp.flush()
                instance = kim_edn.loads(
                    kim_property_create(
                        instance_id=1,
                        property_name=tmp.name,
                    )
                )[0]
        return instance

    @classmethod
    def from_definition(
        cls,
        definitions,
        configuration,
        property_map,
        standardize_energy=False,
        strict=False,
    ):
        """
        A function for constructing a Property given a property setting hash, a property
        definition, and a property map.

        Args:

            definition (dict):
                A valid KIM Property Definition

            configuration (AtomicConfiguration):
                An AtomicConfiguration object from which to extract the property data

            property_map (dict):
                A property map as described in the Property attributes section.

        """
        pdef_dict = {pdef["property-name"]: pdef for pdef in definitions}
        instances = {
            pdef_name: cls.get_kim_instance(pdef)
            for pdef_name, pdef in pdef_dict.items()
        }
        props_dict = {}
        pi_md = None
        # print (property_map.items())
        for pname, pmap_list in property_map.items():
            #print (pname,pmap_list)
            instance = instances.get(pname, None)
            #print ('pname',pname)
            if pname == "_metadata":
                pi_md = md_from_map(pmap_list, configuration)
                #props_dict["method"] = method
                #props_dict["software"] = software
            elif instance is None:
                raise PropertyParsingError(f"Property {pname} not found in definitions")
            else:
                #p_info = MAIN_KEY_MAP.get(pname, None)
            #    if p_info is None:
                    #print(f"property {pname} not found in MAIN_KEY_MAP")
            #        continue
                instance = instance.copy()
                #print (instance)
                not_present = 0
                for pmap_i, pmap in enumerate(pmap_list):
                    #print (pmap.items())
                    #print (configuration.info)
                    for key, val in pmap.items():
                        if "value" in val:
                            # Default value provided
                            data = val["value"]
                        elif val["field"] in configuration.info:
                            data = configuration.info[val["field"]]
                        elif val["field"] in configuration.arrays:
                            data = configuration.arrays[val["field"]]
                        else:
                            # TODO: Populate non existing ones with None
                            # Key not found on configurations. Will be checked later
                            # data = None
                            not_present = 1
                            missing_name = key
                            continue

                        if isinstance(data, (np.ndarray, list)):
                            data = np.atleast_1d(data).tolist()
                        elif isinstance(data, np.integer):
                            data = int(data)
                        elif isinstance(data, np.floating):
                            data = float(data)
                        elif isinstance(data, (str, bool, int, float)):
                            pass
                        instance[key] = {
                            "source-value": data,
                        }

                        if (val["units"] != "None") and (val["units"] is not None):
                            instance[key]["source-unit"] = val["units"]
                # TODO: Check below
                #print ('instance', instance)
                if not_present:
                    if strict:
                        raise Exception(f'Ingestion is set to strict, but some configurations are missing {missing_name}')
                    #print(f"Property {missing_name} not found in {pname}")
                    else:
                        pdef_dict.pop(pname)
                        continue
                # hack to get around OpenKIM requiring the property-name be a dict
                prop_name_tmp = pdef_dict[pname].pop("property-name")

                #print ('instance', instance)
                check_instance_optional_key_marked_required_are_present(
                    instance, pdef_dict[pname]
                )
                pdef_dict[pname]["property-name"] = prop_name_tmp
                props_dict[pname] = {k: v for k, v in instance.items()}
        props_dict["chemical_formula_hill"] = configuration.get_chemical_formula()
        props_dict["configuration_id"] = configuration.id
        return cls(
            definitions=definitions,
            property_map=property_map,
            instance=props_dict,
            metadata=pi_md,
            dataset_id=configuration.dataset_id,
            standardize_energy=standardize_energy,
            nsites=configuration.spark_row["nsites"],
        )

    def to_spark_row(self):
        """
        Convert the Property to a Spark Row object
        """
        # TODO: contruct empty dict from schema or just use instance instead
        row_dict = {}
        #row_dict = _empty_dict_from_schema(property_object_schema)
        row_dict['metadata'] = self.metadata
        for key, val in self.instance.items():
            if key == "method":
                row_dict["method"] = val
            elif key == "software":
                row_dict["software"] = val
            elif key == "_metadata":
                continue
            elif key == "configuration_id":
                row_dict["configuration_id"] = val
            else:
                for k2, v2 in val.items():
                    if k2 in ['property-id', 'instance-id']:
                        continue
                    else:
                        row_dict[f'{key}_{k2}'.replace('-','_')] = v2['source-value']
            '''
            if key == "method":
                row_dict["method"] = val
            elif key == "software":
                row_dict["software"] = val
            elif key == "_metadata":
                continue
            elif key == "configuration_id":
                row_dict["configuration_id"] = val
            elif "energy" in key:
                row_dict.update(prop_to_row_mapper["energy"](key, val))
            else:
                row_dict.update(prop_to_row_mapper[key](val))
            '''
        row_dict["last_modified"] = dateutil.parser.parse(
            datetime.datetime.now(tz=datetime.timezone.utc).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
        )
        row_dict["chemical_formula_hill"] = self.chemical_formula_hill
        row_dict["multiplicity"] = 1
        row_dict["dataset_id"] = self.dataset_id
        return row_dict

    def standardize_energy(self):
        """
        For each key in :attr:`self.property_map`, convert :attr:`self.edn[key]`
        from its original units to the expected ColabFit-compliant units.
        """
        for prop_name, prop_dict in self.instance.items():
            if prop_name not in MAIN_KEY_MAP.keys():
                continue
            p_info = MAIN_KEY_MAP[prop_name]
            if p_info.key not in prop_dict:
                print(f"Property {p_info.key} not found in {prop_name}")
                continue
            units = prop_dict[p_info.key]["source-unit"]
            if p_info.dtype == list:
                prop_val = np.array(
                    prop_dict[p_info.key]["source-value"], dtype=np.float64
                )
            else:
                prop_val = prop_dict[p_info.key]["source-value"]
            if "reference-energy" in prop_dict:
                if prop_dict["reference-energy"]["source-unit"] != units:
                    raise RuntimeError(
                        "Units of the reference energy and energy must be the same"
                    )
                else:
                    prop_val += prop_dict["reference-energy"]["source-value"]

            if "per-atom" in prop_dict:
                if prop_dict["per-atom"]["source-value"] is True:
                    if self.nsites is None:
                        raise RuntimeError("nsites must be provided to convert per-atom")
                    prop_val *= self.nsites

            if units != p_info.unit:
                split_units = list(
                    itertools.chain.from_iterable(
                        [sp.split("/") for sp in units.split("*")]
                    )
                )
                powers = []
                for i, sp_un in enumerate(split_units):
                    if "^" in sp_un:
                        split = sp_un.split("^")
                        powers.append(int(split[1]))
                        split_units[i] = split[0]
                    else:
                        powers.append(1)
                if powers[0] != 1:
                    prop_val *= np.power(float(UNITS[split_units[0]]), powers[0])
                else:
                    prop_val *= float(UNITS[split_units[0]])
                for u, power in zip(split_units[1:], powers[1:]):
                    un = UNITS[u]
                    if power != 1:
                        un = np.power(un, power)
                    if units[units.find(u) - 1] == "*":
                        prop_val *= un
                    elif units[units.find(u) - 1] == "/":
                        prop_val /= un
                    else:
                        raise RuntimeError(
                            f"There may be something wrong with the units: {u}"
                        )
            if p_info.dtype == list:
                prop_val = prop_val.tolist()
            self.instance[prop_name][p_info.key] = {
                "source-value": prop_val,
                "source-unit": p_info.unit,
            }

    def __hash__(self):

        return _hash(
            self.spark_row,
            sorted(self.unique_identifier_kw),
        )

    def __eq__(self, other):
        """
        Returns False if any of the following conditions are true:

        - Properties point to settings with different calculation methods
        - Properties point to different configurations
        - OpenKIM EDN fields differ in any way

        Note that comparison is performed by hashing
        """

        return hash(self) == hash(other)

    def todict(self):
        return self.instance

    def keys(self):
        """Overloaded dictionary function for getting the keys of :attr:`self.edn`"""
        return self.instance.keys()

    def __setitem__(self, k, v):
        """Overloaded :meth:`dict.__setitem__` for setting the values of
        :attr:`self.edn`"""
        edn_key = EDN_KEY_MAP.get(k, k)

        if k in self.instance:
            self.instance[k] = v
        elif edn_key in self.instance:
            self.instance[edn_key] = v
        else:
            KeyError(f"Field '{k}' not found in Property.edn. Returning None")

    def __getitem__(self, k):
        """Overloaded :meth:`dict.__getitem__` for getting the values of
        :attr:`self.edn`"""
        edn_key = EDN_KEY_MAP.get(k, k)

        if k in self.instance:
            return self.instance[k]
        elif edn_key in self.instance:
            return self.instance[edn_key]
        else:
            warnings.warn(f"Field '{k}' not found in Property.edn. Returning None")

            return None

    def get_data(self, k):
        """
        First checks if :code:`k` is in :code:`self.edn`. If not, checks under
        possible pseudonyms. If nothing exists, returns :code:`np.nan`

        Returns:
            data (np.array or np.nan):
                :attr:`self[k]['source-value']` if :code:`k` is a valid key,
                else :code:`np.nan`.
        """
        edn_key = EDN_KEY_MAP.get(k, k)

        if k in self.instance:
            return np.atleast_1d(self[k]["source-value"])
        elif edn_key in self.instance:
            return np.atleast_1d(self[edn_key]["source-value"])
        else:
            return np.nan

    def __delitem__(self, k):
        edn_key = EDN_KEY_MAP.get(k, k)
        del self.instance[edn_key]

    def __str__(self):
        return "Property(properties={})".format(
            [pdef["property-name"] for pdef in self.definitions],
        )

    def __repr__(self):
        return str(self)


class PropertyHashError(Exception):
    pass



import tqdm
import click

OKGREEN = '\033[92m'
END = '\x1b[0m'
@click.group('colabfit')
@click.pass_context
def colabfit(ctx):
    """
    ColabFit Exchange: Data for Advanced Materials and Chemistry
    """
    ctx.ensure_object(dict)



#specify options for query 
#print summary message
#dump info out to stdout and have hyperlinked
#put default values
@colabfit.command('query')
@click.option('--cred', '-c',
        help="""Credential file that contains information about the PostgreSQL host, port, user, database name, etc.
        \b
        """,
        type=click.STRING)
@click.option('--text','-t',
        help="""Perform text search over dataset names, descriptions, authors, and uploader. Full string over substring queries are preferred. 

        \b 
        """,
        type=click.STRING,
        )
@click.option('--elements', '-e',
        help="""Returns datasets that contain specified elements. Datasets may contain more elements in addition to those queried. Use --elements-exact to return datasets that contain ONLY queried elements. Pass arguements as space delimited string.

        \b
        """,
        type=click.STRING)
@click.option('--elements-exact', '-ee',
        help="""Returns datasets that contain ONLY specified elements. Pass arguments as space-delimited string.

        \b

        """,
        type=click.STRING)
@click.option('--property-types','-p',
        help="""Returns datasets that contain queried properties. 

        \b
        See property-name \033]8;;https://github.com/colabfit/colabfit-tools/blob/master/colabfit/tools/property_definitions.py\033\\here\033]8;;\033\\ for relevant properties. Pass arguements as space delimited string.

        \b

        """,
        type=click.STRING)
@click.pass_context
#more
def query(ctx,cred,text,elements,elements_exact,property_types):
    """
    Queries the ColabFit Exchange and prints results. Optionally downloads resulting datasets.
    """
    from .utils import _query, format_print
    import json
    with open(cred) as f:
        d = json.load(f)
    dbname = d['database_name']
    user = d["database_user"]
    port = d["database_port"]
    host = d["database_path"]
    password = d["database_password"]
    q = _query(dbname, user, password, host, port, text,elements,elements_exact,property_types)
    q = list(q)
    print (OKGREEN+"Found %s Dataset(s) in the ColabFit Database!" %len(q)+END)
    print ("--------------------------------------------")
    for i in q:
         format_print(i)
         print ("--------------------------------------------"*2)




from pprint import pprint
import json
import os
import psycopg
from psycopg.rows import dict_row

def _query(dbname, user, password, host, port, text=None, elements=None, elements_exact=None, properties=None):

    query_conditions = []

    if text is not None:
        # Full-text search condition
        query_conditions.append(f"to_tsvector('english', name || ' ' || description || ' ' || array_to_string(authors, ' ') || ' ' || uploader) @@ to_tsquery('english', '{text}')")

    if elements is not None:
        if elements_exact is not None:
            raise Exception('Only one of elements or elements-exact should be specified')
        else:
            # Check if elements array contains all specified elements
            formatted_elements = ", ".join([f"'{e.capitalize()}'" for e in elements.split(' ')])
            query_conditions.append(f"elements @> ARRAY[{formatted_elements}]::VARCHAR[]")

    if elements_exact is not None:
        # Check if elements array matches size and contains exactly the specified elements
        ee = [e.upper() for e in elements_exact.split(' ')]
        formatted_elements_exact = ", ".join([f"'{e}'" for e in ee])
        query_conditions.append(
            f"array_length(elements, 1) = {len(ee)} AND elements @> ARRAY[{formatted_elements_exact}]::VARCHAR[]"
        )

    if properties is not None:
        # Check if property_types array contains all specified properties
        formatted_properties = properties.split(' ')
        for p in formatted_properties:
            query_conditions.append(f"'{p}' = ANY(available_properties)")

    # Combine conditions into a single SQL WHERE clause
    where_clause = " AND ".join(query_conditions)
    if where_clause:
        sql_query = f"SELECT * FROM datasets WHERE {where_clause} ORDER by last_modified;"
    else:    
        sql_query = f"SELECT * FROM datasets ORDER by last_modified;"


    with psycopg.connect(dbname=dbname, user=user, password=password , port=port, host=host, row_factory=dict_row) as conn:
            with conn.cursor() as curs:
                 curs.execute(sql_query)
                 return curs.fetchall()

    
def format_print(doc):
    new_doc={}
    new_doc['colabfit-id']=doc['id']
    new_doc['name']=doc['name']
    new_doc['authors']=doc['authors']
    new_doc['description']=doc['description']
    new_doc['elements']=doc['elements']  
    new_doc['nconfigurations']=doc['nconfigurations']
    new_doc['natoms']=doc['nsites']
    new_doc['available_properties'] = doc['available_properties']
    new_doc['uploader'] = doc['uploader']
    new_doc['last_modified'] = doc['last_modified']
    pprint (new_doc,sort_dicts=False)



import datetime
from collections import defaultdict
import hashlib
import json
import itertools
import os
import string
from ast import literal_eval
from functools import partial
from itertools import islice
from multiprocessing import Pool
from pathlib import Path
from time import time
from types import GeneratorType
from typing import List

import boto3
import dateutil.parser
import psycopg
from psycopg.rows import dict_row
#import pyarrow as pa
#import pyspark.sql.functions as sf
from botocore.exceptions import ClientError
from django.utils.crypto import get_random_string
#from dotenv import load_dotenv
#from ibis import _
from pyspark.sql import Row, SparkSession
from pyspark.sql.types import (
    ArrayType,
    IntegerType,
    LongType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from tqdm import tqdm
#from vastdb.session import Session
from ase import Atoms

from colabfit import (
    ID_FORMAT_STRING,
)  # ATOMS_NAME_FIELD,; EXTENDED_ID_STRING_NAME,; MAX_STRING_LENGTH,; SHORT_ID_STRING_NAME,; _CONFIGS_COLLECTION,; _CONFIGSETS_COLLECTION,; _DATASETS_COLLECTION,; _PROPOBJECT_COLLECTION, # noqa
from colabfit.tools.configuration import AtomicConfiguration
from colabfit.tools.configuration_set import ConfigurationSet
from colabfit.tools.dataset import Dataset
from colabfit.tools.property import Property
from colabfit.tools.property_definitions import atomic_forces_pd, energy_pd, cauchy_stress_pd, quests_descriptor_pd 
from pyspark.sql.functions import udf
from pyspark.sql.types import StringType
from colabfit.tools.schema import (
    config_df_schema,
    config_md_schema,
    config_schema,
    configuration_set_df_schema,
    configuration_set_schema,
    dataset_df_schema,
    dataset_schema,
    property_object_df_schema,
    property_object_md_schema,
    property_object_schema,
    co_cs_mapping_schema,
)
from colabfit.tools.utilities import (
    _hash,
    get_spark_field_type,
    spark_schema_to_arrow_schema,
    split_long_string_cols,
    stringify_df_val,
    unstring_df_val,
)

VAST_BUCKET_DIR = "colabfit-data"
VAST_METADATA_DIR = "data/MD"
NSITES_COL_SPLITS = 20
_CONFIGS_COLLECTION = "test_configs"
_CONFIGSETS_COLLECTION = "test_config_sets"
_DATASETS_COLLECTION = "test_datasets"
_PROPOBJECT_COLLECTION = "test_prop_objects"
_CO_CS_MAP_COLLECTION = "test_co_cs_map"
_MAX_STRING_LEN = 60000

# from kim_property.definition import PROPERTY_ID as VALID_KIM_ID

from kim_property.definition import check_property_definition


def generate_string():
    return get_random_string(12, allowed_chars=string.ascii_lowercase + "1234567890")

'''
class VastDataLoader:
    def __init__(
        self,
        table_prefix: str = "ndb.colabfit.dev",
        endpoint=None,
        access_key=None,
        access_secret=None,
    ):
        self.table_prefix = table_prefix
        self.spark = (
            SparkSession.builder.appName("ColabFitDataLoader")
            # .config("spark.dynamicAllocation.enabled", "true")
            # .config("spark.dynamicAllocation.minExecutors", "1")
            # .config("spark.dynamicAllocation.maxExecutors", "10")
            # .config("spark.task.maxFailures", "4")
            # .config("spark.network.timeout", "300s")
            # .config("spark.speculation", "true")
            # .config("spark.executor.memory", "4g")
            .getOrCreate()
        )
        self.spark.sparkContext.setLogLevel("ERROR")
        if endpoint and access_key and access_secret:
            self.endpoint = endpoint
            self.access_key = access_key
            self.access_secret = access_secret
            self.session = self.get_vastdb_session(
                endpoint=self.endpoint,
                access_key=self.access_key,
                access_secret=self.access_secret,
            )
        self.config_table = f"{self.table_prefix}.{_CONFIGS_COLLECTION}"
        self.config_set_table = f"{self.table_prefix}.{_CONFIGSETS_COLLECTION}"
        self.dataset_table = f"{self.table_prefix}.{_DATASETS_COLLECTION}"
        self.prop_object_table = f"{self.table_prefix}.{_PROPOBJECT_COLLECTION}"
        self.co_cs_map_table = f"{self.table_prefix}.{_CO_CS_MAP_COLLECTION}"

        self.bucket_dir = VAST_BUCKET_DIR
        self.metadata_dir = VAST_METADATA_DIR

    def get_vastdb_session(self, endpoint, access_key: str, access_secret: str):
        return Session(endpoint=endpoint, access=access_key, secret=access_secret)

    def set_vastdb_session(self, endpoint, access_key: str, access_secret: str):
        self.session = self.get_vastdb_session(endpoint, access_key, access_secret)
        self.access_key = access_key
        self.access_secret = access_secret
        self.endpoint = endpoint

    def _get_table_split(self, table_name_str: str):
        """Get bucket, schema and table names for VastDB SDK, with no backticks"""
        table_split = table_name_str.split(".")
        bucket_name = table_split[1].replace("`", "")
        schema_name = table_split[2].replace("`", "")
        table_name = table_split[3].replace("`", "")
        return (bucket_name, schema_name, table_name)

    def add_elem_to_col(df, col_name: str, elem: str):
        df_added_elem = df.withColumn(
            col_name,
            sf.when(
                sf.col(col_name).isNull(), sf.array().cast(ArrayType(StringType()))
            ).otherwise(sf.col(col_name)),
        )
        df_added_elem = df_added_elem.withColumn(
            col_name, sf.array_union(sf.col(col_name), sf.array(sf.lit(elem)))
        )
        return df_added_elem

    def delete_from_table(self, table_name: str, ids: list[str]):
        if isinstance(ids, str):
            ids = [ids]
        bucket_name, schema_name, table_n = self._get_table_split(table_name)
        with self.session.transaction() as tx:
            table = tx.bucket(bucket_name).schema(schema_name).table(table_n)
            rec_batch = table.select(
                predicate=table["id"].isin(ids), internal_row_id=True
            )
            for batch in rec_batch:
                table.delete(rows=batch)

    def check_unique_ids(self, table_name: str, df):
        if not self.spark.catalog.tableExists(table_name):
            print(f"Table {table_name} does not yet exist.")
            return True
        ids = [x["id"] for x in df.select("id").collect()]
        # table_df = self.read_table(table_name)
        # table_df = table_df.select("id")
        bucket_name, schema_name, table_n = self._get_table_split(table_name)
        with self.session.transaction() as tx:
            table = tx.bucket(bucket_name).schema(schema_name).table(table_n)
            for id_batch in tqdm(
                batched(ids, 10000), desc=f"Checking for duplicates in {table_name}"
            ):
                rec_batch_reader = table.select(
                    predicate=table["id"].isin(id_batch), columns=["id"]
                )
                for batch in rec_batch_reader:
                    if batch.num_rows > 0:
                        print(f"Duplicate IDs found in table {table_name}")
                        return False
        return True

        # dupes_exist = id_df.join(table_df, on="id", how="inner")
        # if not dupes_exist.rdd.isEmpty():
        #     # if len(dupes_exist.take(1)) > 0:
        #     print(f"Duplicate IDs found in table {table_name}")
        #     return False
        # return True

    def write_table(
        self,
        spark_df,
        table_name: str,
        ids_filter: list[str] = None,
        check_length_col: str = None,
        check_unique: bool = True,
    ):
        # print(spark_df.first())
        """Include self.table_prefix in the table name when passed to this function"""
        string_schema_dict = {
            self.config_table: config_schema,
            self.config_set_table: configuration_set_schema,
            self.dataset_table: dataset_schema,
            self.prop_object_table: property_object_schema,
            self.co_cs_map_table: co_cs_mapping_schema,
        }
        table_schema = string_schema_dict[table_name]
        if ids_filter is not None:
            spark_df = spark_df.filter(sf.col("id").isin(ids_filter))
        if check_unique:
            all_unique = self.check_unique_ids(table_name, spark_df)
            if not all_unique:
                raise ValueError("Duplicate IDs found in table. Not writing.")
        bucket_name, schema_name, table_n = self._get_table_split(table_name)
        string_cols = [
            f.name for f in spark_df.schema if f.dataType.typeName() == "array"
        ]
        string_col_udf = sf.udf(stringify_df_val, StringType())
        for col in string_cols:
            spark_df = spark_df.withColumn(col, string_col_udf(sf.col(col)))
        if check_length_col is not None:
            spark_df = split_long_string_cols(
                spark_df, check_length_col, _MAX_STRING_LEN
            )
        arrow_schema = spark_schema_to_arrow_schema(table_schema)
        # print(arrow_schema)
        for field in arrow_schema:
            field = field.with_nullable(True)
        if not self.spark.catalog.tableExists(table_name):
            print(f"Creating table {table_name}")

            with self.session.transaction() as tx:
                schema = tx.bucket(bucket_name).schema(schema_name)
                schema.create_table(table_n, arrow_schema)
        arrow_rec_batch = pa.table(
            [pa.array(col) for col in zip(*spark_df.collect())],
            # names=spark_df.columns,
            schema=arrow_schema,
        ).to_batches()
        total_rows = 0
        with self.session.transaction() as tx:
            table = tx.bucket(bucket_name).schema(schema_name).table(table_n)
            for rec_batch in arrow_rec_batch:
                len_batch = rec_batch.num_rows
                table.insert(rec_batch)
                total_rows += len_batch
        print(f"Inserted {total_rows} rows into table {table_name}")

    def write_metadata(self, df):
        """Writes metadata to files using boto3 for VastDB
        Returns a DataFrame without metadata column. The returned DataFrame should
        match table schema (from schema.py)
        """
        if df.filter(sf.col("metadata").isNotNull()).count() == 0:
            df = df.drop("metadata")
            return df
        config = {
            "bucket_dir": self.bucket_dir,
            "access_key": self.access_key,
            "access_secret": self.access_secret,
            "endpoint": self.endpoint,
            "metadata_dir": self.metadata_dir,
        }
        beg = time()
        distinct_metadata = df.select("metadata", "metadata_path").distinct()
        distinct_metadata.foreachPartition(
            lambda partition: write_md_partition(partition, config)
        )
        print(f"Time to write metadata: {time() - beg}")
        df = df.drop("metadata")
        # file_base = f"/vdev/{VAST_BUCKET_DIR}/{VAST_METADATA_DIR}/"
        file_base = f"{self.metadata_dir}/"
        df = df.withColumn(
            "metadata_path",
            prepend_path_udf(sf.lit(str(Path(file_base))), sf.col("metadata_path")),
        )
        return df

    def update_existing_co_po_rows(
        self,
        df,
        table_name,
        cols: list[str],
        elems: list[str],
        str_schema,
        unstr_schema,
        arrow_schema,
        update_cols,
        arr_cols,
    ):
        """
        Updates existing rows in CO or PO table with data from new ingest.

        Parameters:
        -----------
        df : DataFrame
            The DataFrame containing the new data to be updated.
        table_name : str
            The name of the table to be updated.
        cols : list[str]
            List of column names to be updated.
        elems : list[str]
            List of elements corresponding to the columns to be updated.
        str_schema : Schema
            The stringed schema of the table.
        unstr_schema : Schema
            The unstringed schema of the table.
        arrow_schema : Schema
            The Arrow schema of the columns to be updated.
        update_cols : list[str]
            List of columns to be updated.
        arr_cols : list[str]
            List of columns that contain array data.

        Returns:
        --------
        tuple
            A tuple containing two lists:
            - new_ids: List of IDs that were newly added.
            - existing_ids: List of IDs that were updated.
        """
        if isinstance(cols, str):
            cols = [cols]
        if isinstance(elems, str):
            elems = [elems]

        str_col_types = {
            col: get_spark_field_type(str_schema, col) for col in update_cols
        }
        unstr_col_types = {col: get_spark_field_type(unstr_schema, col) for col in cols}
        addtl_fields = {
            "id": StringType(),
            "last_modified": TimestampType(),
            "$row_id": LongType(),
        }
        str_col_types.update(addtl_fields)
        unstr_col_types.update(addtl_fields)
        str_spark_schema = StructType(
            [StructField(col, str_col_types[col], True) for col in update_cols]
            + [
                StructField("id", StringType(), False),
                StructField("$row_id", IntegerType(), False),
            ]
        )
        total_write_cols = update_cols + ["$row_id"]
        ids = [x["id"] for x in df.select("id").collect()]
        batched_ids = batched(ids, 10000)
        new_ids = []
        existing_ids = []
        bucket_name, schema_name, table_n = self._get_table_split(table_name)
        for id_batch in batched_ids:
            id_batch = list(set(id_batch))
            with self.session.transaction() as tx:
                table = tx.bucket(bucket_name).schema(schema_name).table(table_n)
                rec_batch = table.select(
                    predicate=table["id"].isin(id_batch),
                    columns=update_cols + ["id"],
                    internal_row_id=True,
                )
                rec_batch = rec_batch.read_all()
                duplicate_df = self.spark.createDataFrame(
                    rec_batch.to_struct_array().to_pandas(), schema=str_spark_schema
                )
            if duplicate_df.count() == 0:
                new_ids.extend(id_batch)
                continue
            for col_name in arr_cols:
                unstring_udf = sf.udf(unstring_df_val, unstr_col_types[col_name])
                duplicate_df = duplicate_df.withColumn(
                    col_name, unstring_udf(sf.col(col_name))
                )
            for col, elem in zip(cols, elems):
                if col in ["labels", "names"]:
                    if (
                        col == "labels"
                        and df.filter(sf.col("labels").isNotNull()).count() == 0
                    ):
                        continue
                    df_add = df.select("id", col)
                    duplicate_df = (
                        duplicate_df.withColumnRenamed(col, f"{col}_dup")
                        .join(df_add, on="id")
                        .withColumn(
                            col, sf.array_distinct(sf.array_union(f"{col}_dup", col))
                        )
                        .drop(f"{col}_dup")
                    )
                elif col == "multiplicity":
                    df_add = df.select(
                        "id", sf.col("multiplicity").alias("multiplicity_add")
                    )
                    duplicate_df = duplicate_df.join(df_add, on="id", how="left")
                    duplicate_df = duplicate_df.withColumn(
                        "multiplicity",
                        sf.col("multiplicity") + sf.col("multiplicity_add"),
                    ).drop("multiplicity_add")
                else:
                    print(col, unstr_col_types[col])
                    duplicate_df = duplicate_df.withColumn(
                        col, sf.coalesce(sf.col(col), sf.array())
                    )
                    duplicate_df = duplicate_df.withColumn(
                        col,
                        sf.array_distinct(
                            sf.array_union(sf.col(col), sf.array(sf.lit(elem)))
                        ),
                    )
            existing_ids_batch = [x["id"] for x in duplicate_df.select("id").collect()]
            new_ids_batch = [id for id in id_batch if id not in existing_ids_batch]
            string_udf = sf.udf(stringify_df_val, StringType())
            for col_name in arr_cols:
                duplicate_df = duplicate_df.withColumn(
                    col_name, string_udf(sf.col(col_name))
                )
            update_time = dateutil.parser.parse(
                datetime.datetime.now(tz=datetime.timezone.utc).strftime(
                    "%Y-%m-%dT%H:%M:%SZ"
                )
            )
            duplicate_df = duplicate_df.withColumn(
                "last_modified", sf.lit(update_time).cast("timestamp")
            )
            arrow_schema = pa.schema(
                [arrow_schema.field(col) for col in total_write_cols]
            )
            update_table = pa.table(
                [
                    pa.array(col)
                    for col in zip(*duplicate_df.select(total_write_cols).collect())
                ],
                schema=arrow_schema,
            )
            with self.session.transaction() as tx:
                table = tx.bucket(bucket_name).schema(schema_name).table(table_n)
                table.update(rows=update_table, columns=update_cols)
            new_ids.extend(new_ids_batch)
            existing_ids.extend(existing_ids_batch)
        return (new_ids, list(set(existing_ids)))

    def update_existing_co_rows(self, co_df, cols: list[str], elems: list[str]):
        update_cols = [
            col for col in config_schema.fieldNames() if col not in ["id", "$row_id"]
        ]
        # cols_types = [
        #     (col, dtype)
        #     for col, dtype in zip(
        #         cols, [get_spark_field_type(config_df_schema, col) for col in cols]
        #     )
        # ]
        # arr_cols = [
        #     (col, dtype) for col, dtype in cols_types if dtype.typeName() == "array"
        # ]
        arr_cols = [
            col
            for col in cols
            if get_spark_field_type(config_df_schema, col).typeName() == "array"
        ]
        arrow_schema = spark_schema_to_arrow_schema(config_schema)
        arrow_schema = arrow_schema.append(pa.field("$row_id", pa.uint64()))
        return self.update_existing_co_po_rows(
            df=co_df,
            table_name=self.config_table,
            cols=cols,
            elems=elems,
            str_schema=config_schema,
            unstr_schema=config_df_schema,
            arrow_schema=arrow_schema,
            update_cols=update_cols,
            arr_cols=arr_cols,
        )

    def update_existing_po_rows(self, po_df):
        update_cols = ["multiplicity", "last_modified"]
        arr_cols = []
        return self.update_existing_co_po_rows(
            df=po_df,
            table_name=self.prop_object_table,
            cols=["multiplicity"],
            elems=[None],
            str_schema=property_object_schema,
            unstr_schema=property_object_df_schema,
            arrow_schema=pa.schema(
                [
                    pa.field("id", pa.string()),
                    pa.field("multiplicity", pa.int32()),
                    pa.field("last_modified", pa.timestamp("us")),
                    pa.field("$row_id", pa.uint64()),
                ]
            ),
            update_cols=update_cols,
            arr_cols=arr_cols,
        )

    def read_table(
        self, table_name: str, unstring: bool = False, read_metadata: bool = False
    ):
        """
        Include self.table_prefix in the table name when passed to this function.
        Ex: loader.read_table(loader.config_table, unstring=True)
        Arguments:
            table_name {str} -- Name of the table to read from database
        Keyword Arguments:
            unstring {bool} -- Convert stringified lists to lists (default: {False})
            read_metadata {bool} -- Read metadata from files. If True,
            lists will be also converted from strings (default: {False})
        Returns:
            DataFrame -- Spark DataFrame
        """
        string_schema_dict = {
            self.config_table: config_schema,
            self.config_set_table: configuration_set_schema,
            self.dataset_table: dataset_schema,
            self.prop_object_table: property_object_schema,
            self.co_cs_map_table: co_cs_mapping_schema,
        }
        unstring_schema_dict = {
            self.config_table: config_df_schema,
            self.config_set_table: configuration_set_df_schema,
            self.dataset_table: dataset_df_schema,
            self.prop_object_table: property_object_df_schema,
        }
        md_schema_dict = {
            self.config_table: config_md_schema,
            self.config_set_table: configuration_set_df_schema,
            self.dataset_table: dataset_df_schema,
            self.prop_object_table: property_object_md_schema,
        }
        if table_name in [self.config_set_table, self.dataset_table]:
            read_metadata = False
        df = self.spark.read.table(table_name)
        if unstring or read_metadata:
            schema = unstring_schema_dict[table_name]
            schema_type_dict = {f.name: f.dataType for f in schema}
            string_cols = [f.name for f in schema if f.dataType.typeName() == "array"]
            for col in string_cols:
                string_col_udf = sf.udf(unstring_df_val, schema_type_dict[col])
                df = df.withColumn(col, string_col_udf(sf.col(col)))
        if read_metadata:
            schema = md_schema_dict[table_name]
            config = {
                "bucket_dir": self.bucket_dir,
                "access_key": self.access_key,
                "access_secret": self.access_secret,
                "endpoint": self.endpoint,
                "metadata_dir": self.metadata_dir,
            }
            df = df.rdd.mapPartitions(
                lambda partition: read_md_partition(partition, config)
            ).toDF(schema)
        if not read_metadata and not unstring:
            schema = string_schema_dict[table_name]
        mismatched_cols = [
            x
            for x in [(f.name, f.dataType.typeName()) for f in df.schema]
            if x not in [(f.name, f.dataType.typeName()) for f in schema]
        ]
        if len(mismatched_cols) == 0:
            return df
        else:
            raise ValueError(
                f"Schema mismatch for table {table_name}. "
                f"Mismatched column types in DataFrame: {mismatched_cols}"
            )

    def zero_multiplicity(self, dataset_id):
        """Use to return multiplicity of POs for a given dataset to zero"""
        table_exists = self.spark.catalog.tableExists(self.prop_object_table)
        if not table_exists:
            print(f"Table {self.prop_object_table} does not exist")
            return
        spark_schema = StructType(
            [
                StructField("id", StringType(), False),
                StructField("multiplicity", IntegerType(), True),
                StructField("last_modified", TimestampType(), False),
                StructField("$row_id", IntegerType(), False),
            ]
        )
        with self.session.transaction() as tx:
            table_name = self.prop_object_table
            bucket_name, schema_name, table_n = self._get_table_split(table_name)
            table = tx.bucket(bucket_name).schema(schema_name).table(table_n)
            rec_batches = table.select(
                predicate=(table["dataset_id"] == dataset_id)
                & (table["multiplicity"] > 0),
                columns=["id", "multiplicity", "last_modified"],
                internal_row_id=True,
            )
            for rec_batch in rec_batches:
                df = self.spark.createDataFrame(
                    rec_batch.to_struct_array().to_pandas(), schema=spark_schema
                )
                df = df.withColumn("multiplicity", sf.lit(0))
                print(f"Zeroed {df.count()} property objects")
                update_time = dateutil.parser.parse(
                    datetime.datetime.now(tz=datetime.timezone.utc).strftime(
                        "%Y-%m-%dT%H:%M:%SZ"
                    )
                )
                df = df.withColumn(
                    "last_modified", sf.lit(update_time).cast("timestamp")
                )
                arrow_schema = pa.schema(
                    [
                        pa.field("id", pa.string()),
                        pa.field("multiplicity", pa.int32()),
                        pa.field("last_modified", pa.timestamp("us")),
                        pa.field("$row_id", pa.uint64()),
                    ]
                )
                update_table = pa.table(
                    [pa.array(col) for col in zip(*df.collect())], schema=arrow_schema
                )
                table.update(
                    rows=update_table,
                    columns=["multiplicity", "last_modified"],
                )

    def get_pos_cos_by_filter(
        self,
        po_filter_conditions: list[tuple[str, str, str | int | float | list]] = None,
        co_filter_conditions: list[
            tuple[str, str, str | int | float | list | None]
        ] = None,
    ):
        """
        example filter conditions:
        po_filter_conditions = [("dataset_id", "=", "ds_id1"),
                                ("method", "like", "DFT%")]
        co_filter_conditions = [("nsites", ">", 15),
                                ('labels', 'array_contains', 'label1')]
        """
        po_df = self.read_table(self.prop_object_table, unstring=True)
        po_df = self.get_filtered_table(po_df, po_filter_conditions)
        po_df = po_df.drop("chemical_formula_hill")

        co_df = self.read_table(self.config_table, unstring=True)
        overlap_cols = [col for col in po_df.columns if col in co_df.columns]
        po_df = po_df.select(
            [
                (
                    col
                    if col not in overlap_cols
                    else sf.col(col).alias(f"prop_object_{col}")
                )
                for col in po_df.columns
            ]
        )
        co_df = co_df.select(
            [
                (
                    col
                    if col not in overlap_cols
                    else sf.col(col).alias(f"configuration_{col}")
                )
                for col in co_df.columns
            ]
        )
        co_df = self.get_filtered_table(co_df, co_filter_conditions)
        co_po_df = co_df.join(po_df, on="configuration_id", how="inner")
        return co_po_df

    def simple_sdk_query(self, query_table, predicate, schema, internal_row_id=False):
        bucket_name, schema_name, table_n = self._get_table_split(query_table)
        with self.session.transaction() as tx:
            table = tx.bucket(bucket_name).schema(schema_name).table(table_n)
            rec_batch_reader = table.select(
                predicate=predicate, internal_row_id=internal_row_id
            )
            rec_batch = rec_batch_reader.read_all()
            if rec_batch.num_rows == 0:
                print(f"No records found for given query {predicate}")
                return self.spark.createDataFrame([], schema=schema)
            spark_df = self.spark.createDataFrame(
                rec_batch.to_struct_array().to_pandas(), schema=schema
            )
        return spark_df

    def get_co_cs_mapping(self, cs_id: str):
        """
        Get configuration to configuration set mapping for a given ID.

        Args:
            cs_id (str): Configuration set ID.

        Returns:
            DataFrame or None: Mapping DataFrame if found, else None.

        Notes:
            - Prints message and returns None if mapping table doesn't exist.
            - Prints message and returns None if no records found for the given ID.
        """
        if not self.spark.catalog.tableExists(self.co_cs_map_table):
            print(f"Table {self.co_cs_map_table} does not exist")
            return None
        predicate = _.configuration_set_id == cs_id
        co_cs_map = self.simple_sdk_query(
            self.co_cs_map_table, predicate, co_cs_mapping_schema
        )
        if co_cs_map.count() == 0:
            print(f"No records found for given configuration set id {cs_id}")
            return None
        return co_cs_map

    def dataset_query(
        self,
        dataset_id=None,
        table_name=None,
    ):
        if dataset_id is None:
            raise ValueError("dataset_id must be provided")
        string_schema_dict = {
            self.config_table: config_schema,
            self.config_set_table: configuration_set_schema,
            self.dataset_table: dataset_schema,
            self.prop_object_table: property_object_schema,
        }
        df_schema = string_schema_dict[table_name]
        if table_name == self.config_table:
            predicate = _.dataset_ids.contains(dataset_id)
        elif table_name == self.prop_object_table or table_name == self.config_set_table:
            predicate = _.dataset_id == dataset_id
        spark_df = self.simple_sdk_query(table_name, predicate, df_schema)
        unstring_schema_dict = {
            self.config_table: config_df_schema,
            self.config_set_table: configuration_set_df_schema,
            self.dataset_table: dataset_df_schema,
            self.prop_object_table: property_object_df_schema,
        }
        schema = unstring_schema_dict[table_name]
        schema_type_dict = {f.name: f.dataType for f in schema}
        string_cols = [f.name for f in schema if f.dataType.typeName() == "array"]
        for col in string_cols:
            string_col_udf = sf.udf(unstring_df_val, schema_type_dict[col])
            spark_df = spark_df.withColumn(col, string_col_udf(sf.col(col)))
        return spark_df

    def config_set_query(
        self,
        query_table,
        dataset_id=None,
        name_match=None,
        label_match=None,
        configuration_ids=None,
    ):
        if dataset_id is None:
            raise ValueError("dataset_id must be provided")
        string_schema_dict = {
            self.config_table: config_schema,
            self.config_set_table: configuration_set_schema,
            self.dataset_table: dataset_schema,
            self.prop_object_table: property_object_schema,
        }
        df_schema = string_schema_dict[query_table]
        if query_table == self.config_table:
            if name_match is None and label_match is None:
                predicate = _.dataset_ids.contains(dataset_id)
            if name_match is not None and label_match is not None:
                predicate = (
                    (_.dataset_ids.contains(dataset_id))
                    & (_.names.contains(name_match))
                    & (_.labels.contains(label_match))
                )
            elif name_match is not None:
                predicate = (_.dataset_ids.contains(dataset_id)) & (
                    _.names.contains(name_match)
                )
            else:
                predicate = (_.dataset_ids.contains(dataset_id)) & (
                    _.labels.contains(label_match)
                )
            spark_df = self.simple_sdk_query(query_table, predicate, df_schema)
            return spark_df
        elif query_table == self.prop_object_table:
            if configuration_ids is None:
                predicate = _.dataset_id == dataset_id
                spark_df = self.simple_sdk_query(query_table, predicate, df_schema)
            if configuration_ids is not None and len(configuration_ids) < 10000:
                predicate = (_.dataset_id == dataset_id) & (
                    _.configuration_id.isin(configuration_ids)
                )
                spark_df = self.simple_sdk_query(query_table, predicate, df_schema)
            else:
                config_id_batches = batched(configuration_ids, 10000)
                spark_df = self.spark.createDataFrame([], schema=df_schema)
                for batch in config_id_batches:
                    predicate = (_.dataset_id == dataset_id) & (
                        _.configuration_id.isin(batch)
                    )
                    batch_spark_df = self.simple_sdk_query(
                        query_table, predicate, df_schema
                    )
                    spark_df = spark_df.union(batch_spark_df)
            return spark_df

    def get_filtered_table(
        self,
        df,
        filter_conditions: list[tuple[str, str, str | int | float | list]] | None = None,
    ):
        if filter_conditions is None:
            return df
        for i, (column, operand, condition) in enumerate(filter_conditions):
            if operand == "in":
                df = df.filter(sf.col(column).isin(condition))
            elif operand == "like":
                df = df.filter(sf.col(column).like(condition))
            elif operand == "rlike":
                df = df.filter(sf.col(column).rlike(condition))
            elif operand == "==":
                df = df.filter(sf.col(column) == condition)
            elif operand == "array_contains":
                df = df.filter(sf.array_contains(sf.col(column), condition))
            elif operand == ">":
                df = df.filter(sf.col(column) > condition)
            elif operand == "<":
                df = df.filter(sf.col(column) < condition)
            elif operand == ">=":
                df = df.filter(sf.col(column) >= condition)
            elif operand == "<=":
                df = df.filter(sf.col(column) <= condition)
            else:
                raise ValueError(
                    f"Operand {operand} not implemented in get_pos_cos_filter"
                )
        return df

    def rehash_property_objects(spark_row: Row):
        """
        Rehash property object row after changing values of one or
        more of the columns corresponding to hash_keys defined below.

        """
        hash_keys = [
            "adsorption_energy",
            "atomic_forces",
            "atomization_energy",
            "cauchy_stress",
            "cauchy_stress_volume_normalized",
            "chemical_formula_hill",
            "configuration_id",
            "dataset_id",
            "electronic_band_gap",
            "electronic_band_gap_type",
            "energy",
            "formation_energy",
            "metadata_id",
            "method",
            "software",
        ]
        spark_dict = spark_row.asDict()
        if spark_dict["atomic_forces_01"] is None:
            spark_dict["atomic_forces"] = literal_eval(spark_dict["atomic_forces_00"])
        else:
            spark_dict["atomic_forces"] = list(
                itertools.chain(
                    *[
                        literal_eval(spark_dict[f"atomic_forces_{i:02}"])
                        for i in range(1, 19)
                    ]
                )
            )
        if spark_dict["cauchy_stress"] is not None:
            spark_dict["cauchy_stress"] = literal_eval(spark_dict["cauchy_stress"])
        spark_dict["last_modified"] = dateutil.parser.parse(
            datetime.datetime.now(tz=datetime.timezone.utc).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
        )
        spark_dict["hash"] = _hash(spark_dict, hash_keys, include_keys_in_hash=False)
        if spark_dict["cauchy_stress"] is not None:
            spark_dict["cauchy_stress"] = str(spark_dict["cauchy_stress"])
        id = f'PO_{spark_dict["hash"]}'
        if len(id) > 28:
            id = id[:28]
        spark_dict["id"] = id
        return Row(**{k: v for k, v in spark_dict.items() if k != "atomic_forces"})

    @udf(returnType=StringType())
    def config_structure_hash(spark_row: Row, hash_keys: list[str]):
        """
        Rehash configuration object row after changing values of one or
        more of the columns corresponding to hash_keys defined below.

        """
        spark_dict = spark_row.asDict()
        if spark_dict["positions_01"] is None:
            spark_dict["positions"] = literal_eval(spark_dict["positions_00"])
        else:
            spark_dict["positions"] = list(
                itertools.chain(
                    *[
                        literal_eval(spark_dict[f"positions_{i:02}"])
                        for i in range(1, 19)
                    ]
                )
            )
        spark_dict["last_modified"] = dateutil.parser.parse(
            datetime.datetime.now(tz=datetime.timezone.utc).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
        )
        spark_dict["hash"] = _hash(spark_dict, hash_keys, include_keys_in_hash=False)
        return spark_dict["hash"]

    def stop_spark(self):
        self.spark.stop()


class PGDataLoader:
    """
    Class to load data from files to ColabFit PostgreSQL database
    """

    def __init__(
        self,
        appname="colabfit",
        url="jdbc:postgresql://localhost:5432/colabfit",
        database_name: str = None,
        env="./.env",
        table_prefix: str = None,
    ):
        # self.spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", "true")
        JARFILE = os.environ.get("CLASSPATH")
        self.spark = (
            SparkSession.builder.appName(appname)
            .config("spark.jars", JARFILE)
            .getOrCreate()
        )

        user = os.environ.get("PGS_USER")
        password = os.environ.get("PGS_PASS")
        driver = os.environ.get("PGS_DRIVER")
        self.properties = {
            "user": user,
            "password": password,
            "driver": driver,
        }
        self.url = url
        self.database_name = database_name
        self.table_prefix = table_prefix
        findspark.init()

        self.format = "jdbc"  # for postgres local
        load_dotenv(env)
        self.config_table = _CONFIGS_COLLECTION
        self.config_set_table = _CONFIGSETS_COLLECTION
        self.dataset_table = _DATASETS_COLLECTION
        self.prop_object_table = _PROPOBJECT_COLLECTION

    def read_table(
        self,
    ):
        pass

    def get_spark(self):
        return self.spark

    def get_spark_context(self):
        return self.spark.sparkContext

    def write_table(self, spark_rows: list[dict], table_name: str, schema: StructType):
        df = self.spark.createDataFrame(spark_rows, schema=schema)

        df.write.jdbc(
            url=self.url,
            table=table_name,
            mode="append",
            properties=self.properties,
        )

    def write_metadata(self, df):
        """Should accept a DataFrame with a metadata column,
        write metadata to files, return DataFrame without metadata column"""
        pass

    # def update_co_rows_cs_id(self, co_ids: list[str], cs_id: str):
    #     with psycopg.connect(
    #         """dbname=colabfit user=%s password=%s host=localhost port=5432"""
    #         % (
    #             self.user,
    #             self.password,
    #         )
    #     ) as conn:
    #         cur = conn.execute(
    #             """UPDATE configurations
    #                     SET configuration_set_ids = concat(%s::text, \
    #             rtrim(ltrim(replace(configuration_set_ids,%s,''), '['),']'), %s::text)
    #             """,
    #             (
    #                 "[",
    #                 f", {cs_id}",
    #                 f", {cs_id}]",
    #             ),
    #             # WHERE id = ANY(%s)""",
    #             # (cs_id, co_ids),
    #         )
    #         conn.commit()
'''

def batched(configs, n):
    "Batch data into tuples of length n. The last batch may be shorter."
    if not isinstance(configs, GeneratorType):
        configs = iter(configs)
    while True:
        batch = list(islice(configs, n))
        if len(batch) == 0:
            break
        yield batch


class DataManager:
    def __init__(
        self,
        dbname,
        user,
        port,
        host,
        password=None,
        nprocs: int = 1,
        # configs: list[AtomicConfiguration] = None,
        #prop_defs: list[dict] = [atomic_forces_pd, energy_pd, cauchy_stress_pd, quests_descriptor_pd, mask_selection_pd],
        #prop_map: dict = None,
        # dataset_id=None,
        standardize_energy: bool = False,
        read_write_batch_size=10000,
    ):
        self.dbname = dbname
        self.user = user
        self.port = port
        self.user = user
        self.password = password
        self.host = host
        # self.configs = configs
        #if isinstance(prop_defs, dict):
        #    prop_defs = [prop_defs]
        #self.prop_defs = prop_defs
        self.read_write_batch_size = read_write_batch_size
        #self.prop_map = prop_map
        self.nprocs = nprocs
        # self.dataset_id = dataset_id
        self.standardize_energy = standardize_energy
        #if self.dataset_id is None:
        #     self.dataset_id = generate_ds_id()
        #print("Dataset ID:", self.dataset_id)

    @staticmethod
    def _gather_co_po_rows(
        configs: list[AtomicConfiguration],
        prop_defs: list[dict],
        prop_map: dict,
        dataset_id,
        standardize_energy: bool = True,
        strict: bool = False,
    ):
        """Convert COs and DOs to Spark rows."""
        co_po_rows = []
        #po_schema = self.get_table_schema('property_objects')
        for config in configs:
            config.set_dataset_id(dataset_id)
            # TODO: Add PO schema as input to this method so to_spark_row works better
            property = Property.from_definition(
                definitions=prop_defs,
                configuration=config,
                property_map=prop_map,
                standardize_energy=standardize_energy,
                strict=strict,
            )
            co_po_rows.append(
                (
                    config.spark_row,
                    property.spark_row,
                )
            )
        return co_po_rows

    def gather_co_po_rows_pool(
        self, config_chunks: list[list[AtomicConfiguration]], pool, dataset_id=None, prop_map=None, strict=False,
    ):
        """
        Wrapper for _gather_co_po_rows.
        Convert COs and DOs to Spark rows using multiprocessing Pool.
        Returns a batch of tuples of (configuration_row, property_row).
        """

        if dataset_id is None:
            dataset_id = generate_ds_id()

        part_gather = partial(
            self._gather_co_po_rows,
            prop_defs=self.get_property_definitions(),
            prop_map=prop_map,
            dataset_id=dataset_id,
            standardize_energy=self.standardize_energy,
            strict=strict
        )
        return itertools.chain.from_iterable(pool.map(part_gather, list(config_chunks)))

    def gather_co_po_in_batches(self, configs, dataset_id=None, prop_map=None, strict=False):
        """
        Wrapper function for gather_co_po_rows_pool.
        Yields batches of CO-DO rows, preventing configuration iterator from
        being consumed all at once.
        """
        chunk_size = 1000
        config_chunks = batched(configs, chunk_size)
        with Pool(self.nprocs) as pool:
            while True:
                config_batches = list(islice(config_chunks, self.nprocs))
                if not config_batches:
                    break
                else:
                    yield list(self.gather_co_po_rows_pool(config_batches, pool, dataset_id, prop_map, strict))

    def gather_co_po_in_batches_no_pool(self, prop_map=None):
        """
        Wrapper function for gather_co_po_rows_pool.
        Yields batches of CO-DO rows, preventing configuration iterator from
        being consumed all at once.
        """
        chunk_size = self.read_write_batch_size
        config_chunks = batched(self.configs, chunk_size)
        for chunk in config_chunks:
            yield list(
                self._gather_co_po_rows(
                    self.get_property_definitions(),
                    prop_map,
                    self.dataset_id,
                    chunk,
                    standardize_energy=self.standardize_energy,
                )
            )

    def load_co_po_to_vastdb(self, loader, batching_ingest=False):
        if loader.spark.catalog.tableExists(loader.prop_object_table):
            print("loader.prop_object_table exists")
            if batching_ingest is False:
                pos_with_mult = loader.read_table(loader.prop_object_table)
                pos_with_mult = pos_with_mult.filter(
                    sf.col("dataset_id") == self.dataset_id
                )
                pos_with_mult = pos_with_mult.filter(sf.col("multiplicity") > 0).limit(1)
                if pos_with_mult.count() > 0:
                    raise ValueError(
                        f"POs for dataset with ID {self.dataset_id} already exist in "
                        "database with multiplicity > 0.\nTo continue, set "
                        "multiplicities to 0 with "
                        f'loader.zero_multiplicity("{self.dataset_id}")'
                    )
        if loader.spark.catalog.tableExists(loader.dataset_table):
            dataset_exists = loader.read_table(loader.dataset_table).filter(
                sf.col("id") == self.dataset_id
            )
            if dataset_exists.count() > 0:
                raise ValueError(f"Dataset with ID {self.dataset_id} already exists.")
        co_po_rows = self.gather_co_po_in_batches_no_pool()
        for co_po_batch in tqdm(
            co_po_rows,
            desc="Loading data to database: ",
            unit="batch",
        ):
            co_rows, po_rows = list(zip(*co_po_batch))
            if len(co_rows) == 0:
                continue
            else:
                co_df = loader.spark.createDataFrame(co_rows, schema=config_md_schema)
                po_df = loader.spark.createDataFrame(
                    po_rows, schema=property_object_md_schema
                )
                first_count = co_df.count()
                print("Dropping duplicates from CO dataframe")
                merged_names = co_df.groupBy("id").agg(
                    sf.array_distinct(sf.flatten(sf.collect_list("names"))).alias(
                        "names"
                    )
                )
                co_df = co_df.dropDuplicates(["id"])
                second_count = co_df.count()
                if second_count < first_count:
                    co_df = (
                        co_df.drop("names")
                        .join(merged_names, on="id", how="inner")
                        .select(config_md_schema.fieldNames())
                    )
                print(f"{first_count -second_count} duplicates found in CO dataframe")
                count = po_df.count()
                count_distinct = po_df.select("id").distinct().count()
                if count_distinct < count:
                    print(f"{count - count_distinct} duplicates found in PO dataframe")
                    multiplicity = po_df.groupBy("id").agg(sf.count("*").alias("count"))
                    po_df = po_df.dropDuplicates(["id"])
                    po_df = (
                        po_df.join(multiplicity, on="id", how="inner")
                        .withColumn("multiplicity", sf.col("count"))
                        .drop("count")
                    )
                all_unique_co = loader.check_unique_ids(loader.config_table, co_df)
                all_unique_po = loader.check_unique_ids(loader.prop_object_table, po_df)
                if not all_unique_co:
                    new_co_ids, update_co_ids = loader.update_existing_co_rows(
                        co_df=co_df,
                        cols=["dataset_ids", "names", "labels"],
                        elems=[self.dataset_id, None, None],
                    )
                    print(f"Updated {len(update_co_ids)} rows in {loader.config_table}")
                    if len(new_co_ids) > 0:
                        print(f"Writing {len(new_co_ids)} new rows to table")
                        co_df = loader.write_metadata(co_df)
                        loader.write_table(
                            co_df,
                            loader.config_table,
                            ids_filter=new_co_ids,
                            check_length_col="positions_00",
                            check_unique=False,
                        )
                else:
                    co_df = loader.write_metadata(co_df)
                    loader.write_table(
                        co_df,
                        loader.config_table,
                        check_length_col="positions_00",
                        check_unique=False,
                    )
                    # print(f"Inserted {co_df.count()} rows into {loader.config_table}")

                if not all_unique_po:
                    # print("Sending to update_existing_po_rows")
                    new_po_ids, update_po_ids = loader.update_existing_po_rows(
                        po_df=po_df,
                    )
                    print(
                        f"Updated {len(update_po_ids)} rows in "
                        f"{loader.prop_object_table}"
                    )
                    if len(new_po_ids) > 0:
                        print("Remaining POs unique. Writing new rows to table...")
                        po_df = loader.write_metadata(po_df)
                        loader.write_table(
                            po_df,
                            loader.prop_object_table,
                            ids_filter=new_po_ids,
                            check_length_col="atomic_forces_00",
                            check_unique=False,
                        )
                    # print(
                    #     f"Inserted {len(new_po_ids)} rows into "
                    #     f"{loader.prop_object_table}"
                    # )
                else:
                    print("All POs unique: writing to table...")
                    po_df = loader.write_metadata(po_df)
                    # print("finished writing metadata")
                    loader.write_table(
                        po_df,
                        loader.prop_object_table,
                        check_length_col="atomic_forces_00",
                        check_unique=False,
                    )
                    print(
                        f"Inserted {len(po_rows)} rows into {loader.prop_object_table}"
                    )


    def load_data_to_pg_in_batches(self, loader):
        """Load data to PostgreSQL in batches."""
        co_po_rows = self.gather_co_po_in_batches()

        for co_po_batch in tqdm(
            co_po_rows,
            desc="Loading data to database: ",
            unit="batch",
        ):
            co_rows, po_rows = list(zip(*co_po_batch))
            if len(co_rows) == 0:
                continue
            else:
                loader.write_table(
                    co_rows,
                    loader.config_table,
                    config_schema,
                )
                loader.write_table(
                    po_rows,
                    loader.prop_object_table,
                    property_object_schema,
                )

    def load_data_to_pg_in_batches_no_spark(
        self, 
        configs, 
        dataset_id=None, 
        config_table=None, 
        prop_object_table=None, 
        prop_map=None, 
        parameters: dict=None,
        strict=False
    ):
        """
        Load data to PostgreSQL in batches.
        
        :param configs:
        :param datset_id:
        :param config_table:
        :param prop_object_table:
        :param prop_map:
        :param parameters: Dictionary containing two nested dictionaries called
            `universal` and `code` to represent universal values in the 
            database and code specific inputs for the dataset.
        :param strict:
        """

        co_po_rows = self.gather_co_po_in_batches(configs, dataset_id, prop_map, strict)
        for co_po_batch in tqdm(
            co_po_rows,
            desc="Loading data to database: ",
            unit="batch",
        ):
            co_rows, po_rows = list(zip(*co_po_batch))

            if len(co_rows) == 0:
                continue
            
            # make tuple of tuples for data
            column_headers = tuple(co_rows[0].keys())
            co_values = []
            for co_row in co_rows:
                t = []
                for column in column_headers:
                    val = co_row[column]
                    if column == 'last_modified':
                        val = val.strftime("%Y-%m-%dT%H:%M:%SZ")
                    #if isinstance(val, (list, tuple, dict)):
                    #    print (column, type(val[0]))
                    #    val = str(val)
                    t.append(val)
                t.append(co_row['dataset_ids'][0])
                # t.append(co_row['dataset_ids'][0])
                co_values.append(t)
            sql_co = "INSERT INTO configurations (id, hash, last_modified, dataset_ids, configuration_set_ids, chemical_formula_hill, chemical_formula_reduced, chemical_formula_anonymous, elements, elements_ratios, atomic_numbers, nsites, nelements, nperiodic_dimensions, cell, dimension_types, pbc, names, labels, positions) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s) ON CONFLICT (hash) DO UPDATE SET dataset_ids = array_append(configurations.dataset_ids, %s);"
# TODO: Need to modify dataset.from_pg to properly aggregate values and get data to get two copie

#SET dataset_ids = CASE WHEN NOT (%s = ANY(configurations.dataset_ids)) THEN array_append(configurations.dataset_ids, %s) ELSE configurations.dataset_ids END;"
          
            # TODO: Ensure all columns are present here
            # TODO: get column names from query and ensure len matches values
            columns = list(zip(*self.get_table_schema('property_objects')))[0]
            column_string = ', '.join(list(columns))
            val_string = ', '.join(['%s'] * len(columns))
            po_values = []
            for po_row in po_rows:
                t = []
                for column in columns:
                    #print (column)
                    try:
                        val = po_row[column]
                    except:
                        val = None
                    if column == 'last_modified':
                        val = val.strftime("%Y-%m-%dT%H:%M:%SZ")
                    #if isinstance(val, (list, tuple, dict)):
                    #    print (column, type(val[0]))
                    #    val = str(val)
                    t.append(val)
                po_values.append(t)
            # TODO: get column names from query and ensure len matches values
            sql_po = f"""
                INSERT INTO property_objects ({column_string})
                VALUES ({val_string})
                ON CONFLICT (hash)
                DO UPDATE SET multiplicity = property_objects.multiplicity + 1;

            """

            with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
                with conn.cursor() as curs:
                    curs.executemany(sql_co,co_values)
                    curs.executemany(sql_po,po_values)

        if parameters:
            universal = parameters.get('universal', {})
            code = parameters.get('code', {})
            code = json.dumps(code)

            sql_code = f"""
                INSERT INTO dataset_code_specific_parameters (dataset_id, code_specific_inputs)
                VALUES (%s, %s);
                """
            sql_universal = f"""
                INSERT INTO dataset_universal_parameters (dataset_id, parameter_name, parameter_value)
                VALUES (%s, %s, %s)
                """
            
            family_id = dataset_id.split('_')[1]
            universal_values = []
            for key, value in universal.items():
                universal_values.append([family_id, key, value])

            with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
                with conn.cursor() as curs:
                    curs.execute(sql_code, [family_id, code])
                    curs.executemany(sql_universal, universal_values)

    def create_pg_ds_table(self):
        sql = """
        CREATE TABLE datasets (
        id VARCHAR (256),
        hash VARCHAR (256) PRIMARY KEY,
        name VARCHAR (256),
        last_modified VARCHAR (256),
        nconfigurations INT,
        nproperty_objects INT,
        nsites INT,
        elements VARCHAR (1000) [],
        labels VARCHAR (1000) [],
        nelements INT,
        total_elements_ratio DOUBLE PRECISION [],
        nperiodic_dimensions INT [],
        dimension_types VARCHAR (1000) [],
        energy_count INT,
        energy_mean DOUBLE PRECISION,
        energy_variance DOUBLE PRECISION,
        atomic_forces_count INT,
        cauchy_stress_count INT,
        authors VARCHAR (256) [],
        description VARCHAR (10000),
        extended_id VARCHAR (1000),
        license VARCHAR (256),
        links VARCHAR (1000) [],
        publication_year VARCHAR (256),
        doi VARCHAR (256),
        uploader  VARCHAR (256),
        property_map jsonb,
        available_properties VARCHAR (1000) []
        )
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)

    # currently cf-kit table with some properties removed
    def create_pg_po_table(self):
        sql = """
        CREATE TABLE property_objects (
        id VARCHAR (256),
        hash VARCHAR (256) PRIMARY KEY,
        last_modified VARCHAR (256),
        configuration_id VARCHAR (256),
        dataset_id VARCHAR (256),
        multiplicity INT,
        metadata VARCHAR (10000)
        )
        """
        # Don't need anymore
        '''
        chemical_formula_hill VARCHAR (256),
        energy DOUBLE PRECISION,
        atomic_forces_00 DOUBLE PRECISION [] [],
        atomic_forces_01 DOUBLE PRECISION [] [],
        atomic_forces_02 DOUBLE PRECISION [] [],
        atomic_forces_03 DOUBLE PRECISION [] [],
        atomic_forces_04 DOUBLE PRECISION [] [],
        atomic_forces_05 DOUBLE PRECISION [] [],
        atomic_forces_06 DOUBLE PRECISION [] [],
        atomic_forces_07 DOUBLE PRECISION [] [],
        atomic_forces_08 DOUBLE PRECISION [] [],
        atomic_forces_09 DOUBLE PRECISION [] [],
        atomic_forces_10 DOUBLE PRECISION [] [],
        atomic_forces_11 DOUBLE PRECISION [] [],
        atomic_forces_12 DOUBLE PRECISION [] [],
        atomic_forces_13 DOUBLE PRECISION [] [],
        atomic_forces_14 DOUBLE PRECISION [] [],
        atomic_forces_15 DOUBLE PRECISION [] [],
        atomic_forces_16 DOUBLE PRECISION [] [],
        atomic_forces_17 DOUBLE PRECISION [] [],
        atomic_forces_18 DOUBLE PRECISION [] [],
        atomic_forces_19 DOUBLE PRECISION [] [],
        cauchy_stress DOUBLE PRECISION [] []
        )
        '''

        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)

    def create_pg_co_table(self):
        # TODO: Metadata
        sql = """
        CREATE TABLE configurations (
        id VARCHAR (256),
        hash VARCHAR (256) PRIMARY KEY,
        last_modified VARCHAR (256),
        dataset_ids VARCHAR (256) [],
        configuration_set_ids VARCHAR (256) [],
        chemical_formula_hill VARCHAR (256),
        chemical_formula_reduced VARCHAR (256),
        chemical_formula_anonymous VARCHAR (256),
        elements VARCHAR (256) [],
        elements_ratios DOUBLE PRECISION [],
        atomic_numbers INT [],
        nsites INT,
        nelements INT,
        nperiodic_dimensions INT,
        cell DOUBLE PRECISION [] [],
        dimension_types INT [],
        pbc BOOL[],
        names VARCHAR (256) [],
        labels VARCHAR (256) [],
        positions DOUBLE PRECISION [][]
        )
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)

    def create_pg_pd_table(self):
        sql = """
        CREATE TABLE property_definitions (
        hash VARCHAR (256) PRIMARY KEY,
        last_modified VARCHAR (256),
        definition VARCHAR (10000)
        )
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)

    def create_pg_code_table(self):
        sql = """
        CREATE TABLE dataset_code_specific_parameters (
            id integer GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
            dataset_id character varying NOT NULL,
            code_specific_inputs jsonb NOT NULL,
            "timestamp" timestamp without time zone DEFAULT now() NOT NULL
        );
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)

    def create_pg_univ_table(self):
        sql = """
        CREATE TABLE dataset_universal_parameters (
            id integer GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
            dataset_id character varying NOT NULL,
            parameter_name text NOT NULL,
            parameter_value character varying NOT NULL,
            "timestamp" timestamp without time zone DEFAULT now() NOT NULL
        );

        ALTER TABLE ONLY dataset_universal_parameters
            ADD CONSTRAINT dataset_parameters_parameter_definition_fk FOREIGN KEY (parameter_name) REFERENCES parameter_definition(name);
        """

        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)

    def create_pg_param_def_table(self):
        sql = """
        CREATE TABLE parameter_definition (
            id integer GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
            name text NOT NULL UNIQUE,
            description text,
            data_type integer NOT NULL REFERENCES data_type(datatype_id)
        );

        INSERT INTO parameter_definition (name, description, data_type) VALUES
            ('code', 'Which software is being used for the calculation', 3),
            ('version', 'Version of the code used', 3),
            ('executable_path', 'Absolute path to the binary', 3),
            ('xc', 'Which exchange correlation functional used', 3),
            ('planewave_cutoff', 'Planewave cutoff for the code', 1),
            ('kspacing', 'Kspacing in reciprocal space (Not multiplied by 2pi)', 2),
            ('smearing_value', 'Value for smearing', 2),
            ('energy_convergence', 'Energy convergence value', 2),
            ('force_convergence', 'Force convergence', 2),
            ('spin_mode', 'Type of spin calculation', 3),
            ('pseudopotential_library', 'Pseudopotential library being used in simulation', 3),
            ('ion_relax', 'Ionic optimization', 4),
            ('cell_relax', 'Cell optimization', 4),
            ('mixing_mode', 'Electronic mixing method', 3),
            ('mixing_value', 'Electronic mixing value', 2),
            ('vdw_correction', 'Type of Van der Waals correction', 3),
            ('hubbard_method', 'Methodology for applying Hubbard U corrections (Default to Dudarev)', 3),
            ('hubbard_orbitals', 'Set which l-quantum number the correction is applied', 1),
            ('hubbard_u', 'Hubbard U correction value', 2),
            ('hubbard_j', 'Hubbard J correction value', 2),
            ('diagonalization_method', 'Diagonalization method', 3),
            ('magnetic_moments', 'Magnetic moments for each atom', 3),
            ('ion_optimization_method', 'Method for optimizing ions', 3),
            ('smearing_method', 'Type of smearing', 3);
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)

    def create_pg_data_type_table(self):
        sql = """
        CREATE TABLE data_type (
            datatype_id integer GENERATED ALWAYS AS IDENTITY PRIMARY KEY,
            datatype_name varchar NOT NULL UNIQUE,
            description text
        );

        INSERT INTO data_type (datatype_name, description) VALUES
            ('Integer', 'Whole numbers'),
            ('Float', 'Decimal numbers'),
            ('String', 'Text or characters'),
            ('Bool', 'Boolean (True or False)'),
            ('JSON', 'JSON object');
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)

    def create_pg_tables(self):
        """
        Single function to create all the tables needed for the colabfit
        database.
        """

        self.create_pg_ds_table()
        self.create_pg_co_table()
        self.create_pg_po_table()
        self.create_pg_pd_table()
        self.create_pg_data_type_table()
        self.create_pg_param_def_table()
        self.create_pg_code_table()
        self.create_pg_univ_table()

    def insert_property_definition(self, property_dict):
        # TODO: try except that property_dict must be jsonable

        # check that property dict has correct form
        # hack to get around OpenKIM requiring the property-name be a dict
        tmp_prop = property_dict.copy()
        tmp_prop.pop("property-name")
        check_property_definition(tmp_prop)
        for key in ["property-id", "property-name", "property-title", "property-description"]:
            assert key in property_dict, f"{key} must be a part of the property definition format"
        json_pd = json.dumps(property_dict) 
        last_modified = dateutil.parser.parse(
            datetime.datetime.now(tz=datetime.timezone.utc).strftime(
                "%Y-%m-%dT%H:%M:%SZ"
            )
        )
        md5_hash = hashlib.md5(json_pd.encode()).hexdigest()
        sql = """
            INSERT INTO property_definitions (hash, last_modified, definition)
            VALUES (%s, %s, %s)
            ON CONFLICT (hash)
            DO NOTHING
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql, (md5_hash, last_modified, json_pd))
        # TODO: insert columns into po table
        for key, v in property_dict.items():
            if key in ['property-id', 'property-name', 'property-title', 'property-description',]:
                continue
            else:
                column_name = property_dict['property-name'].replace('-', '_') + f'_{key}'.replace('-', '_')
                if v['type'] == 'float':
                    data_type = "DOUBLE PRECISION"
                elif v['type'] == 'int':
                    data_type = "INT"
                elif v['type'] == 'bool':
                    data_type = "BOOL"
                else:
                    data_type = "VARCHAR (10000)"
                for i in range(len(v['extent'])):
                    data_type += '[]' 
            try:
                self.insert_new_column('property_objects', column_name, data_type)

            except Exception as e:
                print(f"An error occurred: {e}")

    def get_property_definitions(self):
        sql = """
             SELECT definition
             FROM property_definitions;
        """ 
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password, row_factory=dict_row) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)
                defs = curs.fetchall()
                dict_defs = []
                for d in defs:
                    dict_defs.append(json.loads(d['definition']))
                return (dict_defs)

    def insert_data_and_create_datset(self,        
        configs,
        name: str,
        authors: list[str],
        description: str,
        publication_link: str = None,
        data_link: str = None,
        dataset_id: str = None,
        other_links: list[str] = None,
        publication_year: str = None,
        doi: str = None,
        labels: list[str] = None,
        data_license: str = "CC-BY-4.0",
        config_table=None,
        prop_object_table=None,
        prop_map=None,
        parameters: dict=None,
        strict=False,
        ):

        if dataset_id is None:
            dataset_id = generate_ds_id()

        # convert to CF AtomicConfiguration if not already
        converted_configs = []
        for c in configs:
            if isinstance(c, Atoms):
                converted_configs.append(AtomicConfiguration.from_ase(c))
            elif isinstance(c, AtomicConfiguration):
                converted_configs.append(c)
            else:
                raise Exception("Configs must be an instance of either ase.Atoms or AtomicConfiguration")

        self.load_data_to_pg_in_batches_no_spark(
            converted_configs, 
            dataset_id, 
            config_table, 
            prop_object_table, 
            prop_map,
            parameters,
            strict
        )
        self.create_dataset_pg_no_spark(
            name,
            dataset_id,
            authors,
            publication_link,
            data_link,
            description,
            other_links,
            publication_year,
            doi,
            labels,
            data_license,
            property_map = prop_map,
        )
        return dataset_id

    def get_table_schema(self, table_name):
        
        # Query to get the table schema
        query = """
        SELECT 
            column_name, 
            data_type, 
            character_maximum_length, 
            is_nullable 
        FROM information_schema.columns 
        WHERE table_name = %s
        ORDER BY ordinal_position;
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(query, (table_name,))
                schema = curs.fetchall()
                return schema
    
    def get_dataset_property_map(self, dataset_id):
        query = """
             SELECT property_map
             FROM datasets
             WHERE id = %s;
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(query, (dataset_id,))
                result = curs.fetchall()
                return result[0][0] 

    def get_dataset_name_from_id(self, dataset_id):
        query = """
             SELECT name
             FROM datasets
             WHERE id = %s;
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(query, (dataset_id,))
                result = curs.fetchall()
                return result[0][0]

    def create_dataset_pg_no_spark(self,
        name: str,
        dataset_id: str,
        authors: list[str],
        publication_link: str,
        data_link: str,
        description: str,
        other_links: list[str] = None,
        publication_year: str = None,
        doi: str = None,
        labels: list[str] = None,
        data_license: str = "CC-BY-4.0",
        property_map: dict = None,
    ):
        # find cs_ids, co_ids, and pi_ids
        config_df = self.dataset_query_pg(dataset_id, 'configurations')
        prop_df = self.dataset_query_pg(dataset_id, 'property_objects')

        if isinstance(authors, str):
            authors = [authors]

        pd = self.get_property_definitions()
        ds = Dataset(
            name=name,
            authors=authors,
            config_df=config_df,
            prop_df=prop_df,
            publication_link=publication_link,
            data_link=data_link,
            description=description,
            other_links=other_links,
            dataset_id=dataset_id,
            labels=labels,
            doi=doi,
            data_license=data_license,
            configuration_set_ids=None,
            publication_year=publication_year,
            use_pg = True,
            property_definitions = pd
        )
        row = ds.spark_row
        try:
            user = os.getlogin()
        except:
            user = "Unknown"
        sql = """
            INSERT INTO datasets (last_modified, nconfigurations, nproperty_objects, nsites, nelements, elements, total_elements_ratio, nperiodic_dimensions, dimension_types, available_properties, energy_mean, energy_variance, atomic_forces_count, cauchy_stress_count, energy_count, authors, description, license, links, name, publication_year, doi, id, extended_id, hash, labels, property_map, uploader)
            VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,%s, %s, %s, %s, %s)
            ON CONFLICT (hash)
            DO NOTHING
        """
        
        column_headers = tuple(row.keys())
        values = []
        for column in column_headers:
            if column in ['nconfiguration_sets']:
                pass
            else:
                val = row[column]
                if column == 'last_modified':
                    val = val.strftime("%Y-%m-%dT%H:%M:%SZ")
                values.append(val)
                #print ('val here',val)
                #print ('t here',t)
                

        values.append(json.dumps(property_map))
        values.append(user)
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.executemany(sql, [values])

    def insert_new_column(self, table, column_name, data_type):
        sql = f"""
            ALTER TABLE {table}
            ADD COLUMN {column_name} {data_type};
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)


    def update_dataset_pg_no_spark(
        self, 
        configs: List[Atoms], 
        dataset_id: str, 
        parameters: dict,
        prop_map: dict, 
        strict=False,
        description=None,
        authors=None,
    ):
        # Get family id
        family = dataset_id.split('_')[1]
        # Check if the provided parameters match with the stored values.
        if parameters:
            sql_param = f"""SELECT code_specific_inputs 
                FROM dataset_code_specific_parameters 
                WHERE dataset_id = '{family}';
                """
            params = self.general_query(sql_param)
            params = params[0]['code_specific_inputs']
            code = parameters.get('code', None)
            if params != code:
                raise ValueError(
                    f"The provided parameters['code'] does not match what is "
                    f"stored in the database. All data being uploaded must have "
                    f"matching input parameters."
                )
        # convert to CF AtomicConfiguration if not already
        converted_configs = []
        for c in configs:
            if isinstance(c, Atoms):
                converted_configs.append(AtomicConfiguration.from_ase(c))
            elif isinstance(c, AtomicConfiguration):
                converted_configs.append(c)
            else:
                raise Exception("Configs must be an instance of either ase.Atoms or AtomicConfiguration")
        # update dataset_id
        q = f"SELECT id FROM datasets where id LIKE '%{family}%'"
        res = self.general_query(q)
        largest_version = sorted([int(r['id'].split('_')[-1]) for r in res])[-1]
        new_v_no = int(largest_version) + 1
        new_dataset_id = dataset_id.split('_')[0] + '_' + dataset_id.split('_')[1] + '_' + str(new_v_no)
        
        self.load_data_to_pg_in_batches_no_spark(converted_configs, new_dataset_id, prop_map=prop_map, strict=strict)

        #config_df_1 = self.dataset_query_pg(dataset_id, 'configurations')
        #prop_df_1 = self.dataset_query_pg(dataset_id, 'property_objects')
        
        config_df_2 = self.dataset_query_pg(new_dataset_id, 'configurations')
        prop_df_2 = self.dataset_query_pg(new_dataset_id, 'property_objects')

        #config_df_1.extend(config_df_2)
        #prop_df_1.extend(prop_df_2)
        
        old_ds = self.get_dataset_pg(dataset_id)[0]

        # format links
        s = old_ds['links'][0].split(' ')[-1].replace("'","")
        d = old_ds['links'][1].split(' ')[-1].replace("'","")
        o = old_ds['links'][2].split(' ')[-1].replace("'","")
        
        if description is None:
            description = old_ds['description']
        if authors is None:
            authors = old_ds['authors']


        pd = self.get_property_definitions()
        ds = Dataset(
            name=old_ds['name'],
            authors=authors,
            config_df=config_df_2,
            prop_df=prop_df_2,
            publication_link=s,
            data_link=d,
            description=description,
            other_links=o,
            dataset_id=new_dataset_id,
            labels=old_ds['labels'],
            doi=old_ds['doi'],
            data_license=old_ds['license'],
            # TODO handle cs later
            configuration_set_ids=None,
            publication_year=old_ds['publication_year'],
            use_pg = True,
            property_definitions = pd
        )
        row = ds.spark_row

        sql = """
            INSERT INTO datasets (last_modified, nconfigurations, nproperty_objects, nsites, nelements, elements, total_elements_ratio, nperiodic_dimensions, dimension_types, energy_mean, energy_variance, atomic_forces_count, cauchy_stress_count, energy_count, authors, description, license, links, name, publication_year, doi, id, extended_id, hash, labels, property_map, uploader)
            VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s,%s, %s, %s, %s)
            ON CONFLICT (hash)
            DO NOTHING
        """

        column_headers = tuple(row.keys())
        values = []
        t = []
        for column in column_headers:
            if column in ['nconfiguration_sets']:
                pass
            else:
                val = row[column]
                if column == 'last_modified':
                    val = val.strftime("%Y-%m-%dT%H:%M:%SZ")
                values.append(val)
        values.append(json.dumps(prop_map))
        try:
            user = os.getlogin()
        except:
            user = "Unknown"
        values.append(user)
       
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.executemany(sql, [values])
                return new_dataset_id

    def get_dataset_data(self, dataset_id):
        sql=f"""
        SELECT
            c.*,  
            po.* 
        FROM
            (SELECT * FROM configurations WHERE '{dataset_id}' = ANY(dataset_ids)) c
        INNER JOIN
            (SELECT * FROM property_objects WHERE dataset_id = '{dataset_id}') po
        ON
            c.id = po.configuration_id
        ORDER by configuration_id;
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password, row_factory=dict_row) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)
                table = curs.fetchall()
                return table

    def general_query(self, sql):
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password, row_factory=dict_row) as conn:
            with conn.cursor() as curs:
                curs.execute(sql)
                try:
                    return curs.fetchall()
                except:
                    return

    def dataset_query_pg(self,
        dataset_id=None,
        table_name=None,
    ):      
        if table_name == 'configurations':
            sql = f"""
                SELECT *
                FROM {table_name}
                WHERE '{dataset_id}' = ANY(dataset_ids);
            """
        elif table_name == 'property_objects':
            sql = f"""
                SELECT *
                FROM {table_name}
                WHERE dataset_id = '{dataset_id}';
            """
        else:
            raise Exception('Only configurations and property_objects tables are supported')
            
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password, row_factory=dict_row) as conn:
            with conn.cursor() as curs:
                r = curs.execute(sql)
                return curs.fetchall()
            
    def get_dataset_pg(self, dataset_id):
        sql = f"""
                SELECT *
                FROM datasets
                WHERE id = '{dataset_id}';
            """ 
        print (dataset_id)    
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password, row_factory=dict_row) as conn:
            with conn.cursor() as curs:
                r = curs.execute(sql)
                return curs.fetchall()

    def create_configuration_sets(
        self,
        loader,
        name_label_match: list[tuple],
    ):
        """
        Args for name_label_match in order:
        1. String pattern for matching CONFIGURATION NAMES
        2. String pattern for matching CONFIGURATION LABELS
        3. Name for configuration set
        4. Description for configuration set
        """
        dataset_id = self.dataset_id
        config_set_rows = []
        # config_df = loader.read_table(table_name=loader.config_table, unstring=True)
        # config_df = config_df.filter(
        #     sf.array_contains(sf.col("dataset_ids"), self.dataset_id)
        # )
        # .cache()
        # prop_df = loader.read_table(loader.prop_object_table, unstring=True)
        # prop_df = prop_df.filter(sf.col("dataset_id") == self.dataset_id)
        # .cache()
        for i, (names_match, label_match, cs_name, cs_desc) in tqdm(
            enumerate(name_label_match), desc="Creating Configuration Sets"
        ):
            print(
                f"names match: {names_match}, label: {label_match}, "
                f"cs_name: {cs_name}, cs_desc: {cs_desc}"
            )
            if names_match and not label_match:
                # config_set_query = config_df.withColumn(
                #     "names_exploded", sf.explode(sf.col("names"))
                # ).filter(sf.col("names_exploded").rlike(names_match))
                config_set_query_df = loader.config_set_query(
                    query_table=loader.config_table,
                    dataset_id=dataset_id,
                    name_match=names_match,
                )
            # Currently an AND operation on labels: labels col contains x AND y
            if label_match and not names_match:
                # if isinstance(label_match, str):
                #     label_match = [label_match]
                # for label in label_match:
                #     config_set_query = config_set_query.filter(
                #         sf.array_contains(sf.col("labels"), label)
                #     )
                config_set_query_df = loader.config_set_query(
                    query_table=loader.config_table,
                    dataset_id=dataset_id,
                    label_match=label_match,
                )
            if names_match and label_match:
                config_set_query_df = loader.config_set_query(
                    query_table=loader.config_table,
                    dataset_id=dataset_id,
                    name_match=names_match,
                    label_match=label_match,
                )
            co_id_df = (
                config_set_query_df.select("id")
                .distinct()
                .withColumnRenamed("id", "configuration_id")
            )
            # prop_df_cs = loader.config_set_query(
            #     query_table=loader.prop_object_table,
            #     dataset_id=dataset_id,
            #     configuration_ids=co_ids,
            # )
            # prop_df_cs = prop_df_cs.select(
            #     "configuration_id", "multiplicity"
            # ).withColumnRenamed("configuration_id", "id")
            # config_set_query_df = config_set_query_df.join(
            #     prop_df_cs, on="id", how="inner"
            # )
            string_cols = [
                "elements",
            ]
            unstring_col_udf = sf.udf(unstring_df_val, ArrayType(StringType()))
            for col in string_cols:
                config_set_query_df = config_set_query_df.withColumn(
                    col, unstring_col_udf(sf.col(col))
                )
            unstring_col_udf = sf.udf(unstring_df_val, ArrayType(IntegerType()))
            int_cols = [
                "atomic_numbers",
                "dimension_types",
            ]
            for col in int_cols:
                config_set_query_df = config_set_query_df.withColumn(
                    col, unstring_col_udf(sf.col(col))
                )
            t = time()
            prelim_cs_id = f"CS_{cs_name}_{self.dataset_id}"
            co_cs_df = loader.get_co_cs_mapping(prelim_cs_id)
            if co_cs_df is not None:
                print(
                    f"Configuration Set {cs_name} already exists.\nRemove rows matching "  # noqa E501
                    f"'configuration_set_id == {prelim_cs_id} from table {loader.co_cs_map_table} to recreate.\n"  # noqa E501
                )
                continue
            config_set = ConfigurationSet(
                name=cs_name,
                description=cs_desc,
                config_df=config_set_query_df,
                dataset_id=self.dataset_id,
            )
            co_cs_df = co_id_df.withColumn("configuration_set_id", sf.lit(config_set.id))
            loader.write_table(co_cs_df, loader.co_cs_map_table, check_unique=False)
            loader.update_existing_co_rows(
                co_df=config_set_query_df,
                cols=["configuration_set_ids"],
                elems=config_set.id,
            )
            t_end = time() - t
            print(f"Time to create CS and update COs with CS-ID: {t_end}")

            config_set_rows.append(config_set.spark_row)
        config_set_df = loader.spark.createDataFrame(
            config_set_rows, schema=configuration_set_df_schema
        )
        loader.write_table(config_set_df, loader.config_set_table)
        return config_set_rows

    def create_dataset(
        self,
        loader,
        name: str,
        authors: list[str],
        publication_link: str,
        data_link: str,
        description: str,
        other_links: list[str] = None,
        publication_year: str = None,
        doi: str = None,
        labels: list[str] = None,
        data_license: str = "CC-BY-4.0",
    ):
        if loader.spark.catalog.tableExists(loader.config_set_table):
            cs_ids = (
                loader.dataset_query(
                    dataset_id=self.dataset_id, table_name=loader.config_set_table
                )
                .select("id")
                .collect()
            )
            # cs_ids = (
            #     loader.read_table(loader.config_set_table)
            #     .filter(sf.col("dataset_id") == self.dataset_id)
            #     .select("id")
            #     .collect()
            # )
            if len(cs_ids) == 0:
                cs_ids = None
            else:
                cs_ids = [x["id"] for x in cs_ids]
        else:
            cs_ids = None
        config_df = loader.dataset_query(
            dataset_id=self.dataset_id, table_name=loader.config_table
        )
        # config_df = loader.read_table(loader.config_table, unstring=True)
        # config_df = config_df.filter(
        #     sf.array_contains(sf.col("dataset_ids"), self.dataset_id)
        # )
        prop_df = loader.dataset_query(
            dataset_id=self.dataset_id, table_name=loader.prop_object_table
        )
        # prop_df = loader.read_table(loader.prop_object_table, unstring=True)
        # prop_df = prop_df.filter(sf.col("dataset_id") == self.dataset_id)
        ds = Dataset(
            name=name,
            authors=authors,
            config_df=config_df,
            prop_df=prop_df,
            publication_link=publication_link,
            data_link=data_link,
            description=description,
            other_links=other_links,
            dataset_id=self.dataset_id,
            labels=labels,
            doi=doi,
            data_license=data_license,
            configuration_set_ids=cs_ids,
            publication_year=publication_year,
        )
        ds_df = loader.spark.createDataFrame([ds.spark_row], schema=dataset_df_schema)
        loader.write_table(ds_df, loader.dataset_table)
    
    def delete_dataset(self, dataset_id, delete_children=False):
        # check if user matches original uploader
        sql = """
            SELECT uploader
            FROM datasets
            WHERE id = %s;
        """
        with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
            with conn.cursor() as curs:
                curs.execute(sql, (dataset_id,))
                uploader = curs.fetchall()[0][0]
        try:
            user = os.getlogin()
        except:
            user = "Unknown"
        if uploader == user:
            print (f'Deleting {dataset_id}')
            sql = """
                DELETE
                FROM datasets
                WHERE id = %s;
            """
            # TODO: delete children as well
            with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
                with conn.cursor() as curs:
                    curs.execute(sql, (dataset_id,))
            if delete_children:
                sql1 = '''
                    DELETE FROM property_objects
                    WHERE dataset_id = %s;
                    '''
                sql2 = '''
                    DELETE FROM configurations
                    WHERE dataset_ids = ARRAY[%s]::varchar[];
                    '''
                sql3 = '''
                    UPDATE configurations
                    SET dataset_ids = array_remove(dataset_ids, %s)
                    WHERE dataset_ids @> ARRAY[%s]::varchar[]  
                    AND cardinality(dataset_ids) > 1;
                '''
                with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
                    with conn.cursor() as curs:
                        curs.execute(sql1, (dataset_id,))
                        curs.execute(sql2, (dataset_id,))
                        curs.execute(sql3, (dataset_id, dataset_id))
        else:
            raise Exception(f'Cannot delete dataset. User must match the original uploader, {uploader}.')

    def delete_items(self, item_id_list):
        # function to COs and POs
        ref_key = {'CO': 'configurations',
                'PO': 'property_objects'}

        assert isinstance(item_id_list, list), 'Input should be a list of PO and/or CO IDs'
        groups = defaultdict(list)
        for s in item_id_list:
            key = s[:2]
            groups[key].append(s)
        groups = dict(groups)
        for k,v in groups.items():
            if k not in ref_key:
                raise Warning(f'This function can only delete COs and POs, but found item with prefix {k}!')
            else:
                sql = f'''
                DELETE FROM {ref_key[k]}
                WHERE id IN {tuple(v)};
                '''
                with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
                    with conn.cursor() as curs:
                        curs.execute(sql)

    def migrate_add_available_properties_column(self):
        """
        Migration method to add the available_properties column to existing datasets tables
        and populate it with data from all datasets.
        """
        try:
            # First, check if the column already exists
            check_sql = """
                SELECT column_name 
                FROM information_schema.columns 
                WHERE table_name = 'datasets' 
                AND column_name = 'available_properties';
            """
            
            with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password) as conn:
                with conn.cursor() as curs:
                    curs.execute(check_sql)
                    column_exists = curs.fetchone()
                    
                    if not column_exists:
                        # Add the column if it doesn't exist
                        add_column_sql = """
                            ALTER TABLE datasets 
                            ADD COLUMN available_properties VARCHAR(1000)[];
                        """
                        curs.execute(add_column_sql)
                        print("Added available_properties column to datasets table")
                    else:
                        print("available_properties column already exists")
                    
                    # Now populate the column for all existing datasets
                    # Get all dataset IDs
                    get_datasets_sql = "SELECT id FROM datasets;"
                    curs.execute(get_datasets_sql)
                    dataset_ids = [row[0] for row in curs.fetchall()]
                    
                    print(f"Found {len(dataset_ids)} datasets to update")
                    
                    for dataset_id in dataset_ids:
                        try:
                            # Get property definitions for this dataset
                            pd_sql = """
                                SELECT definition
                                FROM property_definitions;
                            """
                            curs.execute(pd_sql)
                            prop_defs = [json.loads(row[0]) for row in curs.fetchall()]
                            
                            # Get properties for this dataset
                            props = self.dataset_query_pg(dataset_id, 'property_objects') 
                            if props:
                                # Calculate available properties (similar to Dataset.to_spark_row logic)
                                prop_def_map = {}
                                prop_counts = {}
                                
                                for pd in prop_defs:
                                    for k, v in pd.items():
                                        if k not in ['property-id', 'property-name', 'property-title', 'property-description']:
                                            prop_def_map[f"{pd['property-name'].replace('-','_')}_{k.replace('-','_')}"] = pd['property-name']
                                            prop_counts[f"{pd['property-name'].replace('-','_')}_{k.replace('-','_')}"] = 0
                                            break
                                for p in props:
                                    for k in p.keys():
                                        if str(k) in prop_counts:
                                            if p[str(k)] is not None:
                                                prop_counts[str(k)] += 1
                                
                                available_props = []
                                for k, v in prop_counts.items():
                                    if v > 0:
                                        available_props.append(prop_def_map[k])
                                # Update the dataset with available properties
                                update_sql = """
                                    UPDATE datasets 
                                    SET available_properties = %s 
                                    WHERE id = %s;
                                """
                                curs.execute(update_sql, (available_props, dataset_id))
                                print(f"Updated dataset {dataset_id} with {len(available_props)} available properties")
                            else:
                                # No properties found, set empty array
                                update_sql = """
                                    UPDATE datasets 
                                    SET available_properties = %s 
                                    WHERE id = %s;
                                """
                                curs.execute(update_sql, ([], dataset_id))
                                print(f"Updated dataset {dataset_id} with empty available properties")
                                
                        except Exception as e:
                            print(f"Error updating dataset {dataset_id}: {str(e)}")
                            continue
                    
                    conn.commit()
                    print("Migration completed successfully")
                    
        except Exception as e:
            print(f"Migration failed: {str(e)}")
            raise


class S3BatchManager:
    def __init__(self, bucket_name, access_id, secret_key, endpoint_url=None):
        self.bucket_name = bucket_name
        self.access_id = access_id
        self.secret_key = secret_key
        self.endpoint_url = endpoint_url
        self.client = self.get_client()
        self.MAX_BATCH_SIZE = 100

    def get_client(self):
        return boto3.client(
            "s3",
            use_ssl=False,
            endpoint_url=self.endpoint_url,
            aws_access_key_id=self.access_id,
            aws_secret_access_key=self.secret_key,
            region_name="fake-region",
            config=boto3.session.Config(
                signature_version="s3v4", s3={"addressing_style": "path"}
            ),
        )

    def batch_write(self, file_batch):
        results = []
        for key, content in file_batch:
            try:
                self.client.put_object(Bucket=self.bucket_name, Key=key, Body=content)
                results.append((key, None))
            except Exception as e:
                results.append((key, str(e)))
        return results


def write_md_partition(partition, config):
    s3_mgr = S3BatchManager(
        bucket_name=config["bucket_dir"],
        access_id=config["access_key"],
        secret_key=config["access_secret"],
        endpoint_url=config["endpoint"],
    )
    file_batch = []
    for row in partition:
        md_path = Path(config["metadata_dir"]) / row["metadata_path"]
        file_batch.append((str(md_path), row["metadata"]))

        if len(file_batch) >= s3_mgr.MAX_BATCH_SIZE:
            _ = s3_mgr.batch_write(file_batch)
            file_batch = []
    if file_batch:
        _ = s3_mgr.batch_write(file_batch)
    return iter([])


class S3FileManager:
    def __init__(self, bucket_name, access_id, secret_key, endpoint_url=None):
        self.bucket_name = bucket_name
        self.access_id = access_id
        self.secret_key = secret_key
        self.endpoint_url = endpoint_url

    def get_client(self):
        return boto3.client(
            "s3",
            use_ssl=False,
            endpoint_url=self.endpoint_url,
            aws_access_key_id=self.access_id,
            aws_secret_access_key=self.secret_key,
            region_name="fake-region",
            config=boto3.session.Config(
                signature_version="s3v4", s3={"addressing_style": "path"}
            ),
        )

    def write_file(self, content, file_key):
        try:
            client = self.get_client()
            client.put_object(Bucket=self.bucket_name, Key=file_key, Body=content)
            # return (f"/vdev/{self.bucket_name}/{file_key}", sys.getsizeof(content))
        except Exception as e:
            return f"Error: {str(e)}"

    def read_file(self, file_key):
        try:
            client = self.get_client()
            # key = file_key.replace(str(Path("/vdev/colabfit-data")) + "/", "")
            response = client.get_object(Bucket=self.bucket_name, Key=file_key)
            return response["Body"].read().decode("utf-8")
        except Exception as e:
            return f"Error: {str(e)}"


def generate_ds_id():
    # Maybe check to see whether the DS ID already exists?
    ds_id = ID_FORMAT_STRING.format("DS", generate_string(), 0)
    #print("Generated new DS ID:", ds_id)
    return ds_id

'''
@sf.udf(returnType=StringType())
def prepend_path_udf(prefix, md_path):
    try:
        full_path = Path(prefix) / Path(md_path).relative_to("/")
        return str(full_path)
    except ValueError:
        full_path = Path(prefix) / md_path
        return str(full_path)
'''

# def write_md_partition(partition, config):
#     s3_mgr = S3FileManager(
#         bucket_name=config["bucket_dir"],
#         access_id=config["access_key"],
#         secret_key=config["access_secret"],
#         endpoint_url=config["endpoint"],
#     )
#     for row in partition:
#         md_path = Path(config["metadata_dir"]) / row["metadata_path"]
#         if not md_path.exists():
#             s3_mgr.write_file(
#                 row["metadata"],
#                 str(md_path),
#             )
#     return iter([])


def read_md_partition(partition, config):
    s3_mgr = S3FileManager(
        bucket_name=config["bucket_dir"],
        access_id=config["access_key"],
        secret_key=config["access_secret"],
        endpoint_url=config["endpoint"],
    )

    def process_row(row):
        rowdict = row.asDict()
        try:
            rowdict["metadata"] = s3_mgr.read_file(row["metadata_path"])
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                rowdict["metadata"] = None
            else:
                print(f"Error reading {row['metadata_path']}: {str(e)}")
                rowdict["metadata"] = None
        return Row(**rowdict)

    return map(process_row, partition)

'''
def dataset_query_pg(
    dataset_id=None,
    table_name=None,
):
    if table_name == 'configurations':
        sql = f"""
            SELECT *
            FROM {table_name}
            WHERE '{dataset_id}' = ANY(dataset_ids);
        """
    elif table_name == 'property_objects':
        sql = f"""
            SELECT *
            FROM {table_name}
            WHERE dataset_id = '{dataset_id}';
        """
    else:
        raise Exception('Only configurations and property_objects tables are supported')

    with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password,row_factory=dict_row) as conn:
        with conn.cursor() as curs:
            r = curs.execute(sql)
            return curs.fetchall()

def get_dataset_pg(dataset_id):
    sql = f"""
            SELECT *
            FROM datasets
            WHERE id = '{dataset_id}';
        """

    with psycopg.connect(dbname=self.dbname, user=self.user, port=self.port, host=self.host, password=self.password,row_factory=dict_row) as conn:
        with conn.cursor() as curs:
            r = curs.execute(sql)
            return curs.fetchall()
'''





import numpy as np

__all__ = [
    'BaseTransform',
    'AddDivide',
    'PerAtomEnergies',
    'ReshapeForces',
    'Sequential'
]


class BaseTransform:
    """
    A Transform is used for processing raw data before loading it into a
    Dataset. For example for things like subtracting off a reference energy or
    extracting the 6-component version of the cauchy stress from a 3x3 matrix.
    """

    def __init__(self, tform):
        self._tform = tform

    def __call__(self, data, configurations):
        return self._tform(data, configurations)


class SubtractDivide(BaseTransform):
    """Adds a scalar to the data, then divides by a scalar"""
    def __init__(self, sub, div):
        super(SubtractDivide, self).__init__(lambda x, c: (x-sub)/div)

    def __str__(self):
        return 'SubtractDivide'


class PerAtomEnergies(BaseTransform):
    """Divides the energy by the number of atoms"""
    def __init__(self):
        def wrapper(data, configurations=None):
            return data/len(configurations[0])

        super(PerAtomEnergies, self).__init__(wrapper)

    def __str__(self):
        return 'PerAtomEnergies'


class ReshapeForces(BaseTransform):
    """Reshapes forces into an (N, 3) matrix"""
    def __init__(self):
        def reshape(data, configurations=None):
            data = np.array(data)
            n = np.prod(data.shape)//3

            return data.reshape((n, 3))

        super(ReshapeForces, self).__init__(reshape)

    def __str__(self):
        return 'ReshapeForces'


class Sequential(BaseTransform):
    """
    An object used for defining a chain of Transformations to be performed
    sequentially. For example:

    .. code-block:: python

        Sequential(
            PerAtomEnergies(),
            SubtractDivide(sub=<reference_energy>, div=1)
        )
    """
    def __init__(self, *args):
        def wrapper(data, configurations=None):
            for f in args:
                data = f(data, configurations)
            return data

        super(Sequential, self).__init__(wrapper)


import numpy as np
from hashlib import sha512

from colabfit import HASH_LENGTH, HASH_SHIFT, STRING_DTYPE_SPECIFIER
class PropertySettings:
    """
    This class is used to store information useful for reproducing a Property.

    Attributes:

        method (str):
            A short string describing the method used for computing the
            properties (e.g., 'VASP', 'QuantumEspresso', 'experiment', ...)

        description (str):
            A human-readable description of the settings.

        fields (dict):
            A dictionary of additional information. key = name of the field;


            .. code-block:: python

                value = {
                    'source-value': <key_for_extracting_from_configuration>,
                    'source-units': <string_specifying_units> or None
                    }

            For more details onw how to build this dictionary, refer to the
            "property map" description in the documentation, which follows a
            similar structure. These fields will be used to extract data from a
            Configuration object when inserting into the database.

        files (list):
            A list of 2-tuples, where the first value of each tuple is the name
            of a file, and the second value is the contents of the file.

        labels (list):
            A list of strings; generated by parsing `files`, `description`, and
            `method`.
    """

    def __init__(
        self,
        method='',
        description='',
        fields=None,
        files=None,
        labels=None,
    ):

        if files is None:
            files = []

        if files:
            for ftup in files:
                if not isinstance(ftup, tuple):
                    raise RuntimeError(
                        "PropertySettings files should be 2-tuples of "\
                            "(file_name, file_contents)."
                    )

        if labels is None:
            labels = set()
        elif isinstance(labels, str):
            labels = set([labels])
        else:
            labels = set(labels)

        if fields is None:
            fields = {}

        self.method         = method
        self.fields         = fields
        self.files          = files
        self.description    = description
        self.labels         = labels

        self.parse_labels_from_files()


    def parse_labels_from_files(self):
        pass


    def __hash__(self,):
        """
        Hashes method, description, field contents, and file contents.
        Does NOT use the description or the labels for hashing.
        """

        _hash = sha512()

        file_hashes = []

        for k, vdict in self.fields.items():
            _hash.update(k.encode('utf-8'))

            try:
                hashval =  np.round_(
                    np.array(vdict['source-value']), decimals=12
                ).data.tobytes()
            except:
                try:
                    hashval = np.array(
                        vdict['source-value'], dtype=STRING_DTYPE_SPECIFIER
                    ).data.tobytes()
                except:
                    raise PropertySettingsHashError(
                        "Could not hash key {}: {}".format(k, vdict)
                    )

            _hash.update(hashval)
            _hash.update(str(vdict['source-unit']).encode('utf-8'))


        for (fname, contents) in self.files:
            file_hashes.append(_hash.update(contents.encode('utf-8')))

        _hash.update(self.method.encode('utf-8'))
        # _hash.update(self.description.encode('utf-8'))

        return int(str(int(_hash.hexdigest(), 16)-HASH_SHIFT)[:HASH_LENGTH])


    def __eq__(self, other):
        """Equality check compares hashes"""
        return hash(self) == hash(other)


    def __str__(self):
        return "PropertySettings(method='{}', description='{}', "\
            "labels={})".format(
                self.method,
                self.description,
                self.labels,
            )


    def __repr__(self):
        return str(self)


class PropertySettingsHashError(Exception):
    pass
